# Copyright 2025 BrainX Ecosystem Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

# -*- coding: utf-8 -*-


import brainstate
import braintools
import brainunit as u
import jax
import jax.numpy as jnp
import numpy as np
import pytest

import brainevent
from brainevent._csr.test_util import get_csr, vector_csr, matrix_csr, csr_vector, csr_matrix

platform = jax.default_backend()
BINARY_CSRMV_IMPLEMENTATIONS = tuple(brainevent.binary_csrmv_p.available_backends(platform))
FLOAT_CSRMV_IMPLEMENTATIONS = tuple(brainevent.csrmv_p.available_backends(platform))
FLOAT_CSRMM_IMPLEMENTATIONS = tuple(brainevent.csrmm_p.available_backends(platform))


def gen_events(shape, prob=0.5, asbool=True):
    events = brainstate.random.random(shape) < prob
    if not asbool:
        events = jnp.asarray(events, dtype=float)
    return brainevent.BinaryArray(events)


def gen_sparse_matrix(shape, prob=0.2):
    """
    Generate a sparse matrix with the given shape and sparsity probability.
    """
    matrix = np.random.rand(*shape)
    matrix = np.where(matrix < prob, matrix, 0.)
    return jnp.asarray(matrix, dtype=float)


def ones_like(x):
    return jax.tree.map(jnp.ones_like, x)


def _make_float_data(homo_w, shape):
    if homo_w:
        return jnp.asarray(1.5, dtype=jnp.float32)
    return braintools.init.Normal(0.0, 1.0)(shape)


class Test_CSR_BinaryOperator:
    def test_event_homo_bool(self):
        for dat in [1., 2., 3.]:
            mask = (brainstate.random.rand(10, 20) < 0.1).astype(float) * dat
            csr = u.sparse.CSR.fromdense(mask)
            csr = brainevent.CSR((dat, csr.indices, csr.indptr), shape=mask.shape)

            v = brainevent.BinaryArray(brainstate.random.rand(20) < 0.5)
            assert u.math.allclose(
                mask.astype(float) @ v.value.astype(float),
                csr @ v,
            )

            v = brainevent.BinaryArray(brainstate.random.rand(10) < 0.5)
            assert u.math.allclose(
                v.value.astype(float) @ mask.astype(float),
                v @ csr,
            )

            jax.block_until_ready((mask,))

    def test_event_homo_heter(self):
        mat = brainstate.random.rand(10, 20)
        mask = (brainstate.random.rand(10, 20) < 0.1) * mat
        csr = u.sparse.CSR.fromdense(mask)
        csr = brainevent.CSR((csr.data, csr.indices, csr.indptr), shape=mask.shape)

        v = brainevent.BinaryArray(brainstate.random.rand(20) < 0.5)
        assert u.math.allclose(
            mask.astype(float) @ v.value.astype(float),
            csr @ v,
        )

        v = brainevent.BinaryArray(brainstate.random.rand(10) < 0.5)
        assert u.math.allclose(
            v.value.astype(float) @ mask.astype(float),
            v @ csr,
        )

        jax.block_until_ready((mat, mask))

    def test_event_heter_float_as_bool(self):
        mat = brainstate.random.rand(10, 20)
        mask = (mat < 0.1).astype(float) * mat
        csr = u.sparse.CSR.fromdense(mask)
        csr = brainevent.CSR((csr.data, csr.indices, csr.indptr), shape=mask.shape)

        v = brainevent.BinaryArray((brainstate.random.rand(20) < 0.5).astype(float))
        assert u.math.allclose(
            mask.astype(float) @ v.value.astype(float),
            csr @ v,
        )

        v = brainevent.BinaryArray((brainstate.random.rand(10) < 0.5).astype(float))
        assert u.math.allclose(
            v.value.astype(float) @ mask.astype(float),
            v @ csr,
        )

        jax.block_until_ready((mat, mask))


class Test_CSR_FloatVectorOperator:
    @pytest.mark.parametrize('homo_w', [True, False])
    def test_vector_csr(self, homo_w):
        m, n = 20, 40
        x = brainstate.random.rand(m)
        indptr, indices = get_csr(m, n, 0.1)
        data = _make_float_data(homo_w, indices.shape)
        csr = brainevent.CSR((data, indices, indptr), shape=(m, n))

        y = x @ csr
        y_ref = vector_csr(x, csr.data, indices, indptr, (m, n))
        assert jnp.allclose(y, y_ref, rtol=1e-3, atol=1e-3)

        jax.block_until_ready((x, indptr, indices, data, y, y_ref))

    @pytest.mark.parametrize('homo_w', [True, False])
    def test_csr_vector(self, homo_w):
        m, n = 20, 40
        x = brainstate.random.rand(n)
        indptr, indices = get_csr(m, n, 0.1)
        data = _make_float_data(homo_w, indices.shape)
        csr = brainevent.CSR((data, indices, indptr), shape=(m, n))

        y = csr @ x
        y_ref = csr_vector(x, csr.data, indices, indptr, (m, n))
        assert jnp.allclose(y, y_ref, rtol=1e-3, atol=1e-3)

        jax.block_until_ready((x, indptr, indices, data, y, y_ref))

    @pytest.mark.parametrize('homo_w', [True, False])
    def test_vector_csr_vmap_vector(self, homo_w):
        n_batch, m, n = 10, 20, 40
        xs = brainstate.random.rand(n_batch, m)
        indptr, indices = get_csr(m, n, 0.1)
        data = _make_float_data(homo_w, indices.shape)
        csr = brainevent.CSR((data, indices, indptr), shape=(m, n))

        y = brainstate.transform.vmap2(lambda x: x @ csr)(xs)
        y_ref = brainstate.transform.vmap2(lambda x: vector_csr(x, csr.data, indices, indptr, (m, n)))(xs)
        assert jnp.allclose(y, y_ref, rtol=1e-3, atol=1e-3)

        jax.block_until_ready((xs, indptr, indices, data, y, y_ref))

    @pytest.mark.parametrize('homo_w', [True, False])
    @pytest.mark.parametrize('replace', [True, False])
    @pytest.mark.parametrize('transpose', [True, False])
    def test_vjp(self, homo_w, replace, transpose):
        n_in = 20
        n_out = 30
        shape = (n_in, n_out)
        x = brainstate.random.rand(n_in) if transpose else brainstate.random.rand(n_out)

        indptr, indices = get_csr(n_in, n_out, 0.2, replace=replace)
        w = _make_float_data(homo_w, indices.shape)
        csr = brainevent.CSR((w, indices, indptr), shape=shape)

        def f_brainevent(x_arg, w_arg):
            if transpose:
                return (x_arg @ csr.with_data(w_arg)).sum()
            return (csr.with_data(w_arg) @ x_arg).sum()

        r = jax.grad(f_brainevent, argnums=(0, 1))(x, w)

        def f_ref(x_arg, w_arg):
            if transpose:
                return vector_csr(x_arg, w_arg, indices, indptr, shape=shape).sum()
            return csr_vector(x_arg, w_arg, indices, indptr, shape=shape).sum()

        r_ref = jax.grad(f_ref, argnums=(0, 1))(x, w)
        assert jnp.allclose(r[0], r_ref[0], rtol=1e-3, atol=1e-3)
        assert jnp.allclose(r[1], r_ref[1], rtol=1e-3, atol=1e-3)

        jax.block_until_ready((x, indptr, indices, w, r, r_ref))

    @pytest.mark.parametrize('homo_w', [True, False])
    @pytest.mark.parametrize('replace', [True, False])
    @pytest.mark.parametrize('transpose', [True, False])
    def test_jvp(self, homo_w, replace, transpose):
        n_in = 20
        n_out = 30
        shape = (n_in, n_out)
        x = brainstate.random.rand(n_in if transpose else n_out)

        indptr, indices = get_csr(n_in, n_out, 0.1, replace=replace)
        w = _make_float_data(homo_w, indices.shape)
        csr = brainevent.CSR((w, indices, indptr), shape=shape)

        def f_brainevent(x_arg, w_arg):
            if transpose:
                return x_arg @ csr.with_data(w_arg)
            return csr.with_data(w_arg) @ x_arg

        o1, r1 = jax.jvp(f_brainevent, (x, w), (jnp.ones_like(x), jnp.ones_like(w)))

        def f_ref(x_arg, w_arg):
            if transpose:
                return vector_csr(x_arg, w_arg, indices, indptr, shape=shape)
            return csr_vector(x_arg, w_arg, indices, indptr, shape=shape)

        o2, r2 = jax.jvp(f_ref, (x, w), (jnp.ones_like(x), jnp.ones_like(w)))
        assert jnp.allclose(r1, r2, rtol=1e-3, atol=1e-3)
        assert jnp.allclose(o1, o2, rtol=1e-3, atol=1e-3)

        jax.block_until_ready((x, indptr, indices, w, o1, r1, o2, r2))


@pytest.mark.skipif(
    not FLOAT_CSRMM_IMPLEMENTATIONS,
    reason=f'No csrmm implementation on platform={platform}',
)
class Test_CSR_FloatMatrixOperator:
    @pytest.mark.parametrize('homo_w', [True, False])
    def test_matrix_csr(self, homo_w):
        k, m, n = 10, 20, 40
        x = brainstate.random.rand(k, m)
        indptr, indices = get_csr(m, n, 0.1)
        data = _make_float_data(homo_w, indices.shape)
        csr = brainevent.CSR((data, indices, indptr), shape=(m, n))

        y = x @ csr
        y_ref = matrix_csr(x, csr.data, indices, indptr, (m, n))
        assert jnp.allclose(y, y_ref, rtol=1e-3, atol=1e-3)

        jax.block_until_ready((x, indptr, indices, data, y, y_ref))

    @pytest.mark.parametrize('homo_w', [True, False])
    def test_csr_matrix(self, homo_w):
        m, n, k = 20, 40, 10
        x = brainstate.random.rand(n, k)
        indptr, indices = get_csr(m, n, 0.1)
        data = _make_float_data(homo_w, indices.shape)
        csr = brainevent.CSR((data, indices, indptr), shape=(m, n))

        y = csr @ x
        y_ref = csr_matrix(x, csr.data, indices, indptr, (m, n))
        assert jnp.allclose(y, y_ref, rtol=1e-3, atol=1e-3)

        jax.block_until_ready((x, indptr, indices, data, y, y_ref))


class Test_CSC_CSR_Conversion:
    @pytest.mark.parametrize('shape', [(20, 30), (100, 50)])
    @pytest.mark.parametrize('transpose', [True, False])
    def test_todense(self, shape, transpose):
        matrix = gen_sparse_matrix(shape)
        if transpose:
            matrix = matrix.T
        csr = brainevent.CSR.fromdense(matrix)
        csc = csr.T

        out1 = csr.todense()
        out2 = csc.todense().T
        out3 = csr.T.todense().T
        out4 = csc.T.todense()
        assert jnp.allclose(out1, out2)
        assert jnp.allclose(out1, out3)
        assert jnp.allclose(out1, out4)

        jax.block_until_ready((matrix, out1, out2, out3, out4))

    @pytest.mark.parametrize('shape', [(20, 30), (100, 50)])
    def test_csr_vec(self, shape):
        matrix = gen_sparse_matrix(shape)
        csr = brainevent.CSR.fromdense(matrix)
        csc = csr.T

        vector = jnp.asarray(np.random.rand(shape[1]))

        out1 = csr @ vector
        out2 = vector @ csc
        assert jnp.allclose(out1, out2)

        jax.block_until_ready((matrix, vector, out1, out2))

    @pytest.mark.parametrize('shape', [(20, 30), (100, 50)])
    def test_vec_csr(self, shape):
        matrix = gen_sparse_matrix(shape)
        csr = brainevent.CSR.fromdense(matrix)
        csc = csr.T

        vector = jnp.asarray(np.random.rand(shape[0]))

        out1 = vector @ csr
        out2 = csc @ vector
        assert jnp.allclose(out1, out2)

        jax.block_until_ready((matrix, vector, out1, out2))

    @pytest.mark.parametrize('k', [10])
    @pytest.mark.parametrize('shape', [(20, 30), (100, 50)])
    def test_csr_mat(self, k, shape):
        matrix = gen_sparse_matrix(shape)
        csr = brainevent.CSR.fromdense(matrix)
        csc = csr.T

        matrix = jnp.asarray(np.random.rand(shape[1], k))

        out1 = jax.jit(lambda: csr @ matrix)()
        out2 = jax.jit(lambda: (matrix.T @ csc).T)()
        assert jnp.allclose(out1, out2)

        jax.block_until_ready((matrix, out1, out2))

    # TODO: GPU pallas bug
    @pytest.mark.parametrize('k', [10])
    @pytest.mark.parametrize('shape', [(20, 30), (100, 50)])
    def test_mat_csr(self, k, shape):
        matrix = gen_sparse_matrix(shape)
        csr = brainevent.CSR.fromdense(matrix)
        csc = csr.T

        matrix = jnp.asarray(np.random.rand(k, shape[0]))

        out1 = jax.jit(lambda: matrix @ csr)()
        out2 = jax.jit(lambda: (csc @ matrix.T).T)()
        assert jnp.allclose(out1, out2, atol=1e-4, rtol=1e-4)

        jax.block_until_ready((matrix, out1, out2))

    @pytest.mark.parametrize('shape', [(20, 30), (100, 50)])
    def test_csr_vec_event(self, shape):
        matrix = gen_sparse_matrix(shape)
        csr = brainevent.CSR.fromdense(matrix)
        csc = csr.T

        vector = gen_events(shape[1])

        out1 = jax.jit(lambda: csr @ vector)()
        out2 = jax.jit(lambda: vector @ csc)()
        assert jnp.allclose(out1, out2)

        jax.block_until_ready((matrix, out1, out2))

    @pytest.mark.parametrize('shape', [(20, 30), (100, 50)])
    def test_vec_csr_event(self, shape):
        matrix = gen_sparse_matrix(shape)
        csr = brainevent.CSR.fromdense(matrix)
        csc = csr.T

        vector = gen_events(shape[0])

        out1 = jax.jit(lambda: vector @ csr)()
        out2 = jax.jit(lambda: csc @ vector)()
        assert jnp.allclose(out1, out2)

        jax.block_until_ready((matrix, out1, out2))

    @pytest.mark.parametrize('k', [10])
    @pytest.mark.parametrize('shape', [(20, 30), (100, 50)])
    def test_csr_mat_event(self, k, shape):
        matrix = gen_sparse_matrix(shape)
        csr = brainevent.CSR.fromdense(matrix)
        csc = csr.T

        matrix = gen_events([shape[1], k])

        out1 = jax.jit(lambda: csr @ matrix)()
        out2 = jax.jit(lambda: (matrix.value.T @ csc).T)()
        assert jnp.allclose(out1, out2)

        jax.block_until_ready((matrix, out1, out2))

    # TODO: GPU test error: CUDA_ERROR_ILLEGAL_ADDRESS
    @pytest.mark.parametrize('k', [10])
    @pytest.mark.parametrize('shape', [(20, 30), (100, 50)])
    def test_mat_csr_event(self, k, shape):
        matrix = gen_sparse_matrix(shape)
        csr = brainevent.CSR.fromdense(matrix)
        csc = csr.T

        matrix = gen_events([k, shape[0]])

        out1 = jax.jit(lambda: matrix @ csr)()
        out2 = jax.jit(lambda: (csc @ matrix.value.T).T)()
        assert jnp.allclose(out1, out2, atol=1e-4, rtol=1e-4)

        jax.block_until_ready((matrix, out1, out2))


class Test_CSR:
    @pytest.mark.parametrize('shape', [(20, 30), (100, 50)])
    def test_to_coo(self, shape):
        matrix = gen_sparse_matrix(shape)
        csr = brainevent.CSR.fromdense(matrix)
        coo = csr.tocoo()
        dense = coo.todense()

        assert jnp.allclose(matrix, dense)

        jax.block_until_ready((matrix, dense))

    @pytest.mark.parametrize('shape', [(20, 30), (100, 50)])
    def test_todense(self, shape):
        matrix = gen_sparse_matrix(shape)
        csr = brainevent.CSR.fromdense(matrix)
        dense = csr.todense()
        assert jnp.allclose(matrix, dense)

        jax.block_until_ready((matrix, dense))

    # TODO: GPU pallas error: CUDA_ERROR_ILLEGAL_ADDRESS
    @pytest.mark.parametrize('shape', [(200, 300), (100, 50)])
    @pytest.mark.parametrize('k', [10])
    @pytest.mark.parametrize('transpose', [True, False])
    def test_csr_mat_vjp_heter_weight(self, shape, k, transpose):
        matrix = gen_sparse_matrix(shape)
        csr = brainevent.CSR.fromdense(matrix)
        if transpose:
            xs = brainstate.random.randn(shape[0], k)
        else:
            xs = brainstate.random.randn(shape[1], k)

        def f_brainevent(x, w):
            if transpose:
                return (csr.with_data(w).T @ x).sum()
            else:
                return (csr.with_data(w) @ x).sum()

        def f_dense(x, w):
            if transpose:
                return (csr.with_data(w).T.todense() @ x).sum()
            else:
                return (csr.with_data(w).todense() @ x).sum()

        r1, (g00, g01) = jax.jit(lambda: jax.value_and_grad(f_brainevent, argnums=(0, 1))(xs, csr.data))()
        r2, (g10, g11) = jax.jit(lambda: jax.value_and_grad(f_dense, argnums=(0, 1))(xs, csr.data))()

        rtol = 1e-1 if brainstate.environ.get_platform() == 'gpu' else 1e-4
        assert jnp.allclose(r1, r2, rtol=rtol, atol=rtol)
        assert jnp.allclose(g00, g10, rtol=rtol, atol=rtol)
        assert jnp.allclose(g01, g11, rtol=rtol, atol=rtol)

        jax.block_until_ready((xs, r1, r2, g00, g01, g10, g11))

    @pytest.mark.parametrize('shape', [(200, 300), (100, 50)])
    @pytest.mark.parametrize('k', [10])
    @pytest.mark.parametrize('transpose', [True, False])
    def test_csr_mat_vjp_homo_weight(self, shape, k, transpose):
        matrix = gen_sparse_matrix(shape)
        csr = brainevent.CSR.fromdense(matrix)
        csr = brainevent.CSR((1.5, csr.indices, csr.indptr), shape=shape)

        if transpose:
            xs = brainstate.random.randn(shape[0], k)
        else:
            xs = brainstate.random.randn(shape[1], k)

        def f_brainevent(x, w):
            if transpose:
                return (csr.with_data(w).T @ x).sum()
            else:
                return (csr.with_data(w) @ x).sum()

        def f_dense(x, w):
            if transpose:
                return (csr.with_data(w).T.todense() @ x).sum()
            else:
                return (csr.with_data(w).todense() @ x).sum()

        r1, (g00, g01) = jax.jit(lambda: jax.value_and_grad(f_brainevent, argnums=(0, 1))(xs, csr.data))()
        r2, (g10, g11) = jax.jit(lambda: jax.value_and_grad(f_dense, argnums=(0, 1))(xs, csr.data))()

        assert jnp.allclose(r1, r2, rtol=1e-2, atol=1e-2)
        assert jnp.allclose(g00, g10, rtol=1e-2, atol=1e-2)
        assert jnp.allclose(g01, g11, rtol=1e-2, atol=1e-2)

        jax.block_until_ready((xs, r1, r2, g00, g01, g10, g11))

    @pytest.mark.parametrize('shape', [(200, 300), (100, 50)])
    @pytest.mark.parametrize('transpose', [True, False])
    def test_csr_vec_vjp_heter_weight(self, shape, transpose):
        matrix = gen_sparse_matrix(shape)
        csr = brainevent.CSR.fromdense(matrix)
        if transpose:
            xs = brainstate.random.randn(shape[0])
        else:
            xs = brainstate.random.randn(shape[1])

        def f_brainevent(x, w):
            if transpose:
                return (csr.with_data(w).T @ x).sum()
            else:
                return (csr.with_data(w) @ x).sum()

        def f_dense(x, w):
            if transpose:
                return (csr.with_data(w).T.todense() @ x).sum()
            else:
                return (csr.with_data(w).todense() @ x).sum()

        r1, (g00, g01) = jax.jit(lambda: jax.value_and_grad(f_brainevent, argnums=(0, 1))(xs, csr.data))()
        r2, (g10, g11) = jax.jit(lambda: jax.value_and_grad(f_dense, argnums=(0, 1))(xs, csr.data))()

        assert jnp.allclose(r1, r2, rtol=1e-2, atol=1e-2)
        assert jnp.allclose(g00, g10, rtol=1e-2, atol=1e-2)
        assert jnp.allclose(g01, g11, rtol=1e-2, atol=1e-2)

        jax.block_until_ready((xs, r1, r2, g00, g01, g10, g11))

    @pytest.mark.parametrize('shape', [(200, 300), (100, 50)])
    @pytest.mark.parametrize('transpose', [True, False])
    def test_csr_vec_vjp_homo_weight(self, shape, transpose):
        matrix = gen_sparse_matrix(shape)
        csr = brainevent.CSR.fromdense(matrix)
        csr = brainevent.CSR((1.5, csr.indices, csr.indptr), shape=shape)

        if transpose:
            xs = brainstate.random.randn(shape[0])
        else:
            xs = brainstate.random.randn(shape[1])

        def f_brainevent(x, w):
            if transpose:
                return (csr.with_data(w).T @ x).sum()
            else:
                return (csr.with_data(w) @ x).sum()

        def f_dense(x, w):
            if transpose:
                return (csr.with_data(w).T.todense() @ x).sum()
            else:
                return (csr.with_data(w).todense() @ x).sum()

        r1, (g00, g01) = jax.jit(lambda: jax.value_and_grad(f_brainevent, argnums=(0, 1))(xs, csr.data))()
        r2, (g10, g11) = jax.jit(lambda: jax.value_and_grad(f_dense, argnums=(0, 1))(xs, csr.data))()

        tol = 1e-1 if brainstate.environ.get_platform() else 1e-4
        assert jnp.allclose(r1, r2, rtol=tol, atol=tol)
        assert jnp.allclose(g00, g10, rtol=tol, atol=tol)
        assert jnp.allclose(g01, g11, rtol=tol, atol=tol)

        jax.block_until_ready((xs, r1, r2, g00, g01, g10, g11))

    @pytest.mark.parametrize('shape', [(200, 300), (100, 50)])
    @pytest.mark.parametrize('k', [10])
    @pytest.mark.parametrize('transpose', [True, False])
    def test_csr_mat_jvp_heter_weight(self, shape, k, transpose):
        matrix = gen_sparse_matrix(shape)
        csr = brainevent.CSR.fromdense(matrix)
        if transpose:
            xs = brainstate.random.randn(shape[0], k)
        else:
            xs = brainstate.random.randn(shape[1], k)

        def f_brainevent(x, w):
            if transpose:
                return (csr.with_data(w).T @ x).sum()
            else:
                return (csr.with_data(w) @ x).sum()

        def f_dense(x, w):
            if transpose:
                return (csr.with_data(w).T.todense() @ x).sum()
            else:
                return (csr.with_data(w).todense() @ x).sum()

        r1, g1 = jax.jit(lambda: jax.jvp(f_brainevent,
                                         (xs, csr.data),
                                         (ones_like(xs), ones_like(csr.data))))()
        r2, g2 = jax.jit(lambda: jax.jvp(f_dense,
                                         (xs, csr.data),
                                         (ones_like(xs), ones_like(csr.data))))()

        tol = 1e-1 if brainstate.environ.get_platform() else 1e-4
        assert jnp.allclose(r1, r2, rtol=tol, atol=tol)
        assert jnp.allclose(g1, g2, rtol=tol, atol=tol)

        jax.block_until_ready((xs, r1, g1, r2, g2))

    @pytest.mark.parametrize('shape', [(200, 300), (100, 50)])
    @pytest.mark.parametrize('k', [10])
    @pytest.mark.parametrize('transpose', [True, False])
    def test_csr_mat_jvp_homo_weight(self, shape, k, transpose):
        matrix = gen_sparse_matrix(shape)
        csr = brainevent.CSR.fromdense(matrix)
        csr = brainevent.CSR((1.5, csr.indices, csr.indptr), shape=shape)

        if transpose:
            xs = brainstate.random.randn(shape[0], k)
        else:
            xs = brainstate.random.randn(shape[1], k)

        def f_brainevent(x, w):
            if transpose:
                return (csr.with_data(w).T @ x).sum()
            else:
                return (csr.with_data(w) @ x).sum()

        def f_dense(x, w):
            if transpose:
                return (csr.with_data(w).T.todense() @ x).sum()
            else:
                return (csr.with_data(w).todense() @ x).sum()

        r1, g1 = jax.jit(lambda: jax.jvp(f_brainevent,
                                         (xs, csr.data),
                                         (ones_like(xs), ones_like(csr.data))))()
        r2, g2 = jax.jit(lambda: jax.jvp(f_dense,
                                         (xs, csr.data),
                                         (ones_like(xs), ones_like(csr.data))))()

        tol = 1e-1 if brainstate.environ.get_platform() else 1e-4
        assert jnp.allclose(r1, r2, rtol=tol, atol=tol)
        assert jnp.allclose(g1, g2, rtol=tol, atol=tol)

        jax.block_until_ready((xs, r1, g1, r2, g2))

    @pytest.mark.parametrize('shape', [(200, 300), (100, 50)])
    @pytest.mark.parametrize('transpose', [True, False])
    def test_csr_vec_jvp_heter_weight(self, shape, transpose):
        matrix = gen_sparse_matrix(shape)
        csr = brainevent.CSR.fromdense(matrix)
        if transpose:
            xs = brainstate.random.randn(shape[0])
        else:
            xs = brainstate.random.randn(shape[1])

        def f_brainevent(x, w):
            if transpose:
                return (csr.with_data(w).T @ x).sum()
            else:
                return (csr.with_data(w) @ x).sum()

        def f_dense(x, w):
            if transpose:
                return (csr.with_data(w).T.todense() @ x).sum()
            else:
                return (csr.with_data(w).todense() @ x).sum()

        r1, g1 = jax.jit(lambda: jax.jvp(f_brainevent,
                                         (xs, csr.data),
                                         (ones_like(xs), ones_like(csr.data))))()
        r2, g2 = jax.jit(lambda: jax.jvp(f_dense,
                                         (xs, csr.data),
                                         (ones_like(xs), ones_like(csr.data))))()

        assert jnp.allclose(r1, r2, rtol=1e-2, atol=1e-2)
        assert jnp.allclose(g1, g2, rtol=1e-2, atol=1e-2)

        jax.block_until_ready((xs, r1, g1, r2, g2))

    @pytest.mark.parametrize('shape', [(200, 300), (100, 50)])
    @pytest.mark.parametrize('transpose', [True, False])
    def test_csr_vec_jvp_homo_weight(self, shape, transpose):
        matrix = gen_sparse_matrix(shape)
        csr = brainevent.CSR.fromdense(matrix)
        csr = brainevent.CSR((1.5, csr.indices, csr.indptr), shape=shape)

        if transpose:
            xs = brainstate.random.randn(shape[0])
        else:
            xs = brainstate.random.randn(shape[1])

        def f_brainevent(x, w):
            if transpose:
                return (csr.with_data(w).T @ x).sum()
            else:
                return (csr.with_data(w) @ x).sum()

        def f_dense(x, w):
            if transpose:
                return (csr.with_data(w).T.todense() @ x).sum()
            else:
                return (csr.with_data(w).todense() @ x).sum()

        r1, g1 = jax.jit(lambda: jax.jvp(f_brainevent,
                                         (xs, csr.data),
                                         (ones_like(xs), ones_like(csr.data))))()
        r2, g2 = jax.jit(lambda: jax.jvp(f_dense,
                                         (xs, csr.data),
                                         (ones_like(xs), ones_like(csr.data))))()

        print(r1, r2)
        print(g1, g2)
        assert jnp.allclose(r1, r2, rtol=1e-2, atol=1e-2)
        assert jnp.allclose(g1, g2, rtol=1e-2, atol=1e-2)

        jax.block_until_ready((xs, r1, g1, r2, g2))


class Test_CSR_Event:

    @pytest.mark.parametrize('shape', [(200, 300), (100, 50)])
    @pytest.mark.parametrize('k', [10])
    @pytest.mark.parametrize('transpose', [True, False])
    def test_csr_mat_vjp_heter_weight(self, shape, k, transpose):
        matrix = gen_sparse_matrix(shape)
        csr = brainevent.CSR.fromdense(matrix)

        if transpose:
            xs = gen_events([shape[0], k], asbool=False)
        else:
            xs = gen_events([shape[1], k], asbool=False)

        def f_brainevent(x, w):
            if transpose:
                return (csr.with_data(w).T @ x).sum()
            else:
                return (csr.with_data(w) @ x).sum()

        def f_dense(x, w):
            if transpose:
                return (csr.with_data(w).T.todense() @ x).sum()
            else:
                return (csr.with_data(w).todense() @ x).sum()

        r1, (g00, g01) = jax.jit(lambda: jax.value_and_grad(f_brainevent, argnums=(0, 1))(xs, csr.data))()
        r2, (g10, g11) = jax.jit(lambda: jax.value_and_grad(f_dense, argnums=(0, 1))(xs, csr.data))()

        assert isinstance(g00, brainevent.BinaryArray)
        assert isinstance(g10, brainevent.BinaryArray)

        assert jnp.allclose(r1, r2, rtol=1e-2, atol=1e-2)
        assert jnp.allclose(g00.value, g10.value, rtol=1e-2, atol=1e-2)
        assert jnp.allclose(g01, g11, rtol=1e-2, atol=1e-2)

        jax.block_until_ready((r1, r2, g00, g01, g10, g11))

    @pytest.mark.parametrize('shape', [(200, 300), (100, 50)])
    @pytest.mark.parametrize('k', [10])
    @pytest.mark.parametrize('transpose', [True, False])
    def test_csr_mat_vjp_homo_weight(self, shape, k, transpose):
        matrix = gen_sparse_matrix(shape)
        csr = brainevent.CSR.fromdense(matrix)
        csr = brainevent.CSR((1.5, csr.indices, csr.indptr), shape=shape)

        if transpose:
            xs = gen_events([shape[0], k], asbool=False)
        else:
            xs = gen_events([shape[1], k], asbool=False)

        def f_brainevent(x, w):
            if transpose:
                return (csr.with_data(w).T @ x).sum()
            else:
                return (csr.with_data(w) @ x).sum()

        def f_dense(x, w):
            if transpose:
                return (csr.with_data(w).T.todense() @ x).sum()
            else:
                return (csr.with_data(w).todense() @ x).sum()

        r1, (g00, g01) = jax.jit(lambda: jax.value_and_grad(f_brainevent, argnums=(0, 1))(xs, csr.data))()
        r2, (g10, g11) = jax.jit(lambda: jax.value_and_grad(f_dense, argnums=(0, 1))(xs, csr.data))()

        assert isinstance(g00, brainevent.BinaryArray)
        assert isinstance(g10, brainevent.BinaryArray)

        assert jnp.allclose(r1, r2, rtol=1e-2, atol=1e-2)
        assert jnp.allclose(g00.value, g10.value, rtol=1e-2, atol=1e-2)
        assert jnp.allclose(g01, g11, rtol=1e-2, atol=1e-2)

        jax.block_until_ready((r1, r2, g00, g01, g10, g11))

    @pytest.mark.parametrize('shape', [(200, 300), (100, 50)])
    @pytest.mark.parametrize('transpose', [True, False])
    def test_csr_vec_vjp_heter_weight(self, shape, transpose):
        matrix = gen_sparse_matrix(shape)
        csr = brainevent.CSR.fromdense(matrix)

        if transpose:
            xs = gen_events([shape[0], ], asbool=False)
        else:
            xs = gen_events([shape[1], ], asbool=False)

        def f_brainevent(x, w):
            if transpose:
                return (csr.with_data(w).T @ x).sum()
            else:
                return (csr.with_data(w) @ x).sum()

        def f_dense(x, w):
            if transpose:
                return (csr.with_data(w).T.todense() @ x).sum()
            else:
                return (csr.with_data(w).todense() @ x).sum()

        r1, (g00, g01) = jax.jit(lambda: jax.value_and_grad(f_brainevent, argnums=(0, 1))(xs, csr.data))()
        r2, (g10, g11) = jax.jit(lambda: jax.value_and_grad(f_dense, argnums=(0, 1))(xs, csr.data))()

        assert isinstance(g00, brainevent.BinaryArray)
        assert isinstance(g10, brainevent.BinaryArray)

        assert jnp.allclose(r1, r2, rtol=1e-2, atol=1e-2)
        assert jnp.allclose(g00.value, g10.value, rtol=1e-2, atol=1e-2)
        assert jnp.allclose(g01, g11, rtol=1e-2, atol=1e-2)

        jax.block_until_ready((r1, r2, g00, g01, g10, g11))

    @pytest.mark.parametrize('shape', [(200, 300), (100, 50)])
    @pytest.mark.parametrize('transpose', [True, False])
    def test_csr_vec_vjp_homo_weight(self, shape, transpose):
        matrix = gen_sparse_matrix(shape)
        csr = brainevent.CSR.fromdense(matrix)
        csr = brainevent.CSR((1.5, csr.indices, csr.indptr), shape=shape)

        if transpose:
            xs = gen_events([shape[0], ], asbool=False)
        else:
            xs = gen_events([shape[1], ], asbool=False)

        def f_brainevent(x, w):
            if transpose:
                return (csr.with_data(w).T @ x).sum()
            else:
                return (csr.with_data(w) @ x).sum()

        def f_dense(x, w):
            if transpose:
                return (csr.with_data(w).T.todense() @ x).sum()
            else:
                return (csr.with_data(w).todense() @ x).sum()

        r1, (g00, g01) = jax.jit(lambda: jax.value_and_grad(f_brainevent, argnums=(0, 1))(xs, csr.data))()
        r2, (g10, g11) = jax.jit(lambda: jax.value_and_grad(f_dense, argnums=(0, 1))(xs, csr.data))()

        assert isinstance(g00, brainevent.BinaryArray)
        assert isinstance(g10, brainevent.BinaryArray)

        assert jnp.allclose(r1, r2, rtol=1e-2, atol=1e-2)
        assert jnp.allclose(g00.value, g10.value, rtol=1e-2, atol=1e-2)
        assert jnp.allclose(g01, g11, rtol=1e-2, atol=1e-2)

        jax.block_until_ready((r1, r2, g00, g01, g10, g11))

    @pytest.mark.parametrize('shape', [(200, 300), (100, 50)])
    @pytest.mark.parametrize('k', [10])
    @pytest.mark.parametrize('transpose', [True, False])
    def test_csr_mat_jvp_heter_weight(self, shape, k, transpose):
        matrix = gen_sparse_matrix(shape)
        csr = brainevent.CSR.fromdense(matrix)
        if transpose:
            xs = gen_events([shape[0], k], asbool=False)
        else:
            xs = gen_events([shape[1], k], asbool=False)

        def f_brainevent(x, w):
            if transpose:
                return (csr.with_data(w).T @ x).sum()
            else:
                return (csr.with_data(w) @ x).sum()

        def f_dense(x, w):
            if transpose:
                return (csr.with_data(w).T.todense() @ x).sum()
            else:
                return (csr.with_data(w).todense() @ x).sum()

        r1, g1 = jax.jit(lambda: jax.jvp(f_brainevent,
                                         (xs, csr.data),
                                         (ones_like(xs), ones_like(csr.data))))()
        r2, g2 = jax.jit(lambda: jax.jvp(f_dense,
                                         (xs, csr.data),
                                         (ones_like(xs), ones_like(csr.data))))()

        assert jnp.allclose(r1, r2, rtol=1e-2, atol=1e-2)
        assert jnp.allclose(g1, g2, rtol=1e-2, atol=1e-2)

        jax.block_until_ready((r1, g1, r2, g2))

    @pytest.mark.parametrize('shape', [(200, 300), (100, 50)])
    @pytest.mark.parametrize('k', [10])
    @pytest.mark.parametrize('transpose', [True, False])
    def test_csr_mat_jvp_homo_weight(self, shape, k, transpose):
        matrix = gen_sparse_matrix(shape)
        csr = brainevent.CSR.fromdense(matrix)
        csr = brainevent.CSR((1.5, csr.indices, csr.indptr), shape=shape)

        if transpose:
            xs = gen_events([shape[0], k], asbool=False)
        else:
            xs = gen_events([shape[1], k], asbool=False)

        def f_brainevent(x, w):
            if transpose:
                return (csr.with_data(w).T @ x).sum()
            else:
                return (csr.with_data(w) @ x).sum()

        def f_dense(x, w):
            if transpose:
                return (csr.with_data(w).T.todense() @ x).sum()
            else:
                return (csr.with_data(w).todense() @ x).sum()

        r1, g1 = jax.jit(lambda: jax.jvp(f_brainevent,
                                         (xs, csr.data),
                                         (ones_like(xs), ones_like(csr.data))))()
        r2, g2 = jax.jit(lambda: jax.jvp(f_dense,
                                         (xs, csr.data),
                                         (ones_like(xs), ones_like(csr.data))))()

        assert jnp.allclose(r1, r2, rtol=1e-2, atol=1e-2)
        assert jnp.allclose(g1, g2, rtol=1e-2, atol=1e-2)

        jax.block_until_ready((r1, g1, r2, g2))

    @pytest.mark.parametrize('shape', [(200, 300), (100, 50)])
    @pytest.mark.parametrize('transpose', [True, False])
    def test_csr_vec_jvp_heter_weight(self, shape, transpose):
        matrix = gen_sparse_matrix(shape)
        csr = brainevent.CSR.fromdense(matrix)

        if transpose:
            xs = gen_events([shape[0], ], asbool=False)
        else:
            xs = gen_events([shape[1], ], asbool=False)

        def f_brainevent(x, w):
            if transpose:
                return (csr.with_data(w).T @ x).sum()
            else:
                return (csr.with_data(w) @ x).sum()

        def f_dense(x, w):
            if transpose:
                return (csr.with_data(w).T.todense() @ x).sum()
            else:
                return (csr.with_data(w).todense() @ x).sum()

        r1, g1 = jax.jit(lambda: jax.jvp(f_brainevent,
                                         (xs, csr.data),
                                         (ones_like(xs), ones_like(csr.data))))()
        r2, g2 = jax.jit(lambda: jax.jvp(f_dense,
                                         (xs, csr.data),
                                         (ones_like(xs), ones_like(csr.data))))()

        assert jnp.allclose(r1, r2, rtol=1e-2, atol=1e-2)
        assert jnp.allclose(g1, g2, rtol=1e-2, atol=1e-2)

        jax.block_until_ready((r1, g1, r2, g2))

    @pytest.mark.parametrize('shape', [(200, 300), (100, 50)])
    @pytest.mark.parametrize('transpose', [True, False])
    def test_csr_vec_jvp_homo_weight(self, shape, transpose):
        matrix = gen_sparse_matrix(shape)
        csr = brainevent.CSR.fromdense(matrix)
        csr = brainevent.CSR((1.5, csr.indices, csr.indptr), shape=shape)

        if transpose:
            xs = gen_events([shape[0], ], asbool=False)
        else:
            xs = gen_events([shape[1], ], asbool=False)

        def f_brainevent(x, w):
            if transpose:
                return (csr.with_data(w).T @ x).sum()
            else:
                return (csr.with_data(w) @ x).sum()

        def f_dense(x, w):
            if transpose:
                return (csr.with_data(w).T.todense() @ x).sum()
            else:
                return (csr.with_data(w).todense() @ x).sum()

        r1, g1 = jax.jit(lambda: jax.jvp(f_brainevent,
                                         (xs, csr.data),
                                         (ones_like(xs), ones_like(csr.data))))()
        r2, g2 = jax.jit(lambda: jax.jvp(f_dense,
                                         (xs, csr.data),
                                         (ones_like(xs), ones_like(csr.data))))()

        assert jnp.allclose(r1, r2, rtol=1e-2, atol=1e-2)
        assert jnp.allclose(g1, g2, rtol=1e-2, atol=1e-2)

        jax.block_until_ready((r1, g1, r2, g2))


class Test_CSC:
    @pytest.mark.parametrize('shape', [(20, 30), (100, 50)])
    def test_to_coo(self, shape):
        matrix = gen_sparse_matrix(shape)
        csc = brainevent.CSR.fromdense(matrix).T
        coo = csc.tocoo()
        dense = coo.todense()
        assert jnp.allclose(matrix.T, dense)

        jax.block_until_ready((dense,))

    @pytest.mark.parametrize('shape', [(20, 30), (100, 50)])
    def test_todense(self, shape):
        matrix = gen_sparse_matrix(shape)
        csr = brainevent.CSR.fromdense(matrix).T
        dense = csr.todense()
        assert jnp.allclose(matrix.T, dense)

        jax.block_until_ready((dense,))


class Test_diag_add:
    @pytest.mark.parametrize('shape', [(200, 300), (100, 50), (400, 400)])
    def test_csr(self, shape):
        dense = brainstate.random.rand(*shape)
        mask = dense < 0.1
        dense = jnp.where(mask, dense, 0.)
        csr = brainevent.CSR.fromdense(dense)
        diag = brainstate.random.rand(min(shape))
        new_csr = csr.diag_add(diag)

        new_dense = new_csr.todense()
        dense = dense.at[jnp.diag_indices(min(shape))].add(diag)
        dense = jnp.where(mask, dense, 0.)

        print(new_dense)
        print(dense)

        assert jnp.allclose(new_dense, dense)

        jax.block_until_ready((dense, diag, new_dense))

    @pytest.mark.parametrize('shape', [(200, 300), (100, 50), (400, 400)])
    def test_csc(self, shape):
        dense = brainstate.random.rand(*shape)
        mask = dense < 0.1
        dense = jnp.where(mask, dense, 0.)
        csc = brainevent.CSC.fromdense(dense)
        diag = brainstate.random.rand(min(shape))
        new_csr = csc.diag_add(diag)

        new_dense = new_csr.todense()
        dense = dense.at[jnp.diag_indices(min(shape))].add(diag)
        dense = jnp.where(mask, dense, 0.)

        print(new_dense)
        print(dense)

        assert jnp.allclose(new_dense, dense)

        jax.block_until_ready((dense, diag, new_dense))

    @pytest.mark.parametrize('shape', [(200, 300), (100, 50), (400, 400)])
    def test_csr_and_csc(self, shape):
        dense = brainstate.random.rand(*shape)
        mask = dense < 0.1
        dense = jnp.where(mask, dense, 0.)
        csr = brainevent.CSR.fromdense(dense)
        csc = csr.T
        diag = brainstate.random.rand(min(shape))
        new_csr = csc.diag_add(diag)

        new_dense = new_csr.todense().T
        dense = dense.at[jnp.diag_indices(min(shape))].add(diag)
        dense = jnp.where(mask, dense, 0.)

        print(new_dense)
        print(dense)

        assert jnp.allclose(new_dense, dense)

        jax.block_until_ready((dense, diag, new_dense))


class Test_solve:
    @pytest.mark.parametrize('shape', [(200, 200), (400, 400)])
    def test_csr(self, shape: brainstate.typing.Shape):
        dense = brainstate.random.rand(*shape)
        mask = dense < 0.1
        dense = jnp.where(mask, dense, 0.)
        csr = brainevent.CSR.fromdense(dense)
        b = brainstate.random.randn(shape[0])

        x = csr.solve(b)
        assert jnp.allclose(csr @ x, b, atol=1e0, rtol=1e0)

        x2 = jnp.linalg.solve(dense, b)
        assert jnp.allclose(x, x2, atol=1e0, rtol=1e0)

        jax.block_until_ready((dense, b, x, x2))

    @pytest.mark.parametrize('shape', [(200, 200), (400, 400)])
    def test_csc(self, shape: brainstate.typing.Shape):
        dense = brainstate.random.rand(*shape)
        mask = dense < 0.1
        dense = jnp.where(mask, dense, 0.)
        csc = brainevent.CSR.fromdense(dense)
        b = brainstate.random.randn(shape[0])

        x = csc.solve(b)
        assert jnp.allclose(csc @ x, b, atol=1e0, rtol=1e0)

        x2 = jnp.linalg.solve(dense, b)
        assert jnp.allclose(x, x2, atol=1e0, rtol=1e0)

        jax.block_until_ready((dense, b, x, x2))



# Copyright 2025 BrainX Ecosystem Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

# -*- coding: utf-8 -*-

import functools
import importlib.util
from typing import Protocol, Union, Tuple, Sequence

import jax
import numpy as np
from jax import tree_util
from jax.interpreters import ad

from brainevent._compatible_import import Primitive

warp_installed = importlib.util.find_spec('warp') is not None
tvmffi_installed = importlib.util.find_spec('jax_tvm_ffi') is not None

# Try to import TVM FFI - will fail gracefully if not available
if tvmffi_installed:
    try:
        import jax_tvm_ffi
        import tvm_ffi.cpp
    except:
        tvmffi_installed = False

# Try to import Warp - will fail gracefully if not available
if warp_installed:
    try:
        import warp  # pylint: disable=import-error, import-outside-toplevel

        warp.config.quiet = True
    except:
        warp_installed = False

__all__ = [
    'register_tvm_cuda_kernels',
    'defjvp',
    'general_batching_rule',
    'jaxtype_to_warptype',
    'jaxinfo_to_warpinfo',
    'check_pallas_jax_version',
    'check_warp_installed',
]

_MIN_JAX_VERSION_FOR_PALLAS = (0, 7, 1)


def check_pallas_jax_version():
    """Check that the installed JAX version satisfies Pallas requirements.

    Pallas kernels require JAX >= 0.7.1.  This function verifies the
    installed version and raises an informative error if it is too old.

    Raises
    ------
    RuntimeError
        If the installed JAX version is older than 0.7.1.

    Examples
    --------
    .. code-block:: python

        >>> check_pallas_jax_version()  # succeeds silently on JAX >= 0.7.1
    """
    if jax.__version_info__ < _MIN_JAX_VERSION_FOR_PALLAS:
        min_ver = '.'.join(str(v) for v in _MIN_JAX_VERSION_FOR_PALLAS)
        raise RuntimeError(
            f"Pallas kernels require JAX >= {min_ver}, "
            f"but found JAX {jax.__version__}. "
            f"Please upgrade JAX: pip install --upgrade jax"
        )


def check_warp_installed():
    """Check that the Warp package is installed and importable.

    This function verifies that the ``warp`` Python package is available
    at runtime.  It is called automatically before dispatching to a Warp
    kernel backend.

    Raises
    ------
    RuntimeError
        If the ``warp`` package is not installed or could not be
        imported.

    See Also
    --------
    jaxtype_to_warptype : Convert a JAX/NumPy dtype to a Warp type.
    jaxinfo_to_warpinfo : Convert a ``jax.ShapeDtypeStruct`` to a Warp
        array type.

    Examples
    --------
    .. code-block:: python

        >>> check_warp_installed()  # succeeds silently when warp is installed
    """
    if not warp_installed:
        raise RuntimeError(
            "Warp kernels require the 'warp' package, but it is not installed.\n"
            "Please install Warp using one of the following methods:\n"
            "  pip install warp-lang\n"
            "For more information, visit: https://nvidia.github.io/warp/user_guide/installation.html"
        )


def register_tvm_cuda_kernels(
    source_code: str,
    module: str,
    functions: Sequence[str],
):
    """Compile CUDA source code and register the resulting kernels with JAX FFI.

    Uses the TVM FFI infrastructure (``jax_tvm_ffi`` and ``tvm_ffi.cpp``)
    to compile inline CUDA source and register each resulting function as
    a JAX FFI target on the GPU platform.  If the TVM FFI packages are
    not installed the function returns silently, allowing upstream code
    to fall back to other backends.

    Parameters
    ----------
    source_code : str
        CUDA C/C++ source code containing the kernel implementations.
    module : str
        Module name under which the compiled kernels are registered.
        Each kernel is registered as ``"<module>.<function_name>"``.
    functions : Sequence of str
        Names of the functions (entry points) to extract from the
        compiled module and register with JAX FFI.

    Returns
    -------
    None

    Raises
    ------
    ValueError
        If *source_code* is not a string, *module* is not a string, or
        *functions* is not a sequence of strings.  (Note: the current
        implementation returns the ``ValueError`` instead of raising it;
        this may change in a future version.)

    See Also
    --------
    XLACustomKernel.def_tvmffi_kernel : Register a TVM FFI kernel with
        an ``XLACustomKernel``.

    Notes
    -----
    The function requires both ``jax_tvm_ffi`` and ``tvm_ffi.cpp``
    packages to be installed.  If they are not available, the function
    returns silently with no effect.  The compiled kernels are
    registered as JAX FFI targets using the naming convention
    ``"<module>.<function_name>"`` and are available on the ``"gpu"``
    platform.

    Examples
    --------
    .. code-block:: python

        >>> cuda_src = '''
        ... extern "C" void add_arrays(float* a, float* b, float* out, int n) {
        ...     int i = threadIdx.x + blockIdx.x * blockDim.x;
        ...     if (i < n) out[i] = a[i] + b[i];
        ... }
        ... '''
        >>> register_tvm_cuda_kernels(cuda_src, 'my_kernels', ['add_arrays'])
    """

    if not tvmffi_installed:
        return

    if not isinstance(source_code, str):
        return ValueError("source_code must be a string")
    if not isinstance(module, str):
        return ValueError("module must be a string")
    if not isinstance(functions, Sequence) or not all(isinstance(f, str) for f in functions):
        return ValueError("functions must be a sequence of strings")

    # Compile CUDA module
    _cuda_module = tvm_ffi.cpp.load_inline(name=module, cuda_sources=source_code, functions=functions)

    # Register each kernel with JAX FFI
    for name in functions:
        jax_tvm_ffi.register_ffi_target(
            f"{module}.{name}",
            getattr(_cuda_module, name),
            ["args", "rets", "ctx.stream"],
            platform="gpu",
        )


def defjvp(primitive, *jvp_rules):
    """Define per-input JVP rules for a JAX primitive.

    This function allows defining Jacobian-vector product (JVP) rules
    for JAX primitives, extending the functionality of
    ``jax.interpreters.ad.defjvp``.  While the standard JAX function
    primarily supports primitives that return a single result
    (``multiple_results=False``), this implementation supports defining
    independent JVP rules for each input parameter regardless of whether
    the primitive returns single or multiple results.

    This is particularly useful for custom operations or primitives
    where different inputs have different differentiation rules, or where
    the primitive produces multiple outputs that need distinct handling
    in automatic differentiation.

    Parameters
    ----------
    primitive : Primitive or XLACustomKernel
        The JAX ``Primitive`` object (or an ``XLACustomKernel`` instance,
        from which the underlying ``Primitive`` is extracted) for which
        the JVP rule is being defined.
    *jvp_rules : callable or None
        One callable per input primal.  Each callable has the signature
        ``rule(tangent, *primals, **params) -> tangent_out`` and computes
        the tangent contribution from the corresponding input.  Pass
        ``None`` for inputs whose JVP contribution is zero.

    Raises
    ------
    AssertionError
        If *primitive* (after unwrapping) is not a ``Primitive`` instance.

    See Also
    --------
    XLACustomKernel.def_jvp_rule : Register a single monolithic JVP
        rule for an ``XLACustomKernel``.
    XLACustomKernel.def_jvp_rule2 : Convenience wrapper around this
        function.
    general_batching_rule : General batching rule for custom primitives.

    Notes
    -----
    When the primitive has ``multiple_results=True``, a custom internal
    JVP implementation (``_standard_jvp``) is used that correctly
    handles tuple outputs.  For single-result primitives, the standard
    ``jax.interpreters.ad.standard_jvp`` is used.

    Examples
    --------
    .. code-block:: python

        >>> # Assume `my_prim` is a JAX Primitive with two inputs.
        >>> def jvp_rule_input0(tangent, x, y, **kw):
        ...     return tangent * y
        >>> def jvp_rule_input1(tangent, x, y, **kw):
        ...     return tangent * x
        >>> defjvp(my_prim, jvp_rule_input0, jvp_rule_input1)
    """
    # Import XLACustomKernel locally to avoid circular dependencies.
    from .main import XLACustomKernel

    # If the input is an XLACustomKernel, extract the underlying JAX primitive.
    if isinstance(primitive, XLACustomKernel):
        primitive = primitive.primitive
    # Ensure that the 'primitive' argument is indeed a JAX Primitive object.
    assert isinstance(primitive, Primitive), f'The primitive should be a JAX primitive. But we got {primitive}'

    # Check if the primitive returns multiple results.
    if primitive.multiple_results:
        # If yes, use the custom _standard_jvp function designed to handle multiple results.
        # ad.primitive_jvps is the JAX registry for JVP rules.
        # functools.partial pre-fills the jvp_rules and primitive arguments for _standard_jvp.
        ad.primitive_jvps[primitive] = functools.partial(_standard_jvp, jvp_rules, primitive)
    else:
        # If no (single result), use the standard JAX JVP handler (ad.standard_jvp).
        # This maintains compatibility with standard JAX behavior for single-result primitives.
        ad.primitive_jvps[primitive] = functools.partial(ad.standard_jvp, jvp_rules, primitive)


def _standard_jvp(jvp_rules, primitive: Primitive, primals, tangents, **params):
    """Compute the JVP for a multi-result primitive.

    This is an internal helper used by :func:`defjvp` when the primitive
    has ``multiple_results=True``.  It binds the primals, evaluates
    each per-input JVP rule whose tangent is not ``Zero``, and sums the
    resulting tangent contributions.

    Parameters
    ----------
    jvp_rules : sequence of callable or None
        Per-input JVP rules (see :func:`defjvp`).
    primitive : Primitive
        The JAX primitive.
    primals : tuple
        Primal input values.
    tangents : tuple
        Tangent input values (may contain ``ad.Zero``).
    **params
        Additional keyword arguments forwarded from the primitive bind.

    Returns
    -------
    tuple
        A pair ``(val_out, tangents_out)`` where *val_out* is the tuple
        of primal outputs and *tangents_out* is the summed tangent
        contributions with the same pytree structure as *val_out*.
    """
    assert primitive.multiple_results
    val_out = tuple(primitive.bind(*primals, **params))
    tree = tree_util.tree_structure(val_out)
    tangents_out = []
    for rule, t in zip(jvp_rules, tangents):
        if rule is not None and type(t) is not ad.Zero:
            r = tuple(rule(t, *primals, **params))
            tangents_out.append(r)
            assert tree_util.tree_structure(r) == tree
    r = functools.reduce(
        _add_tangents,
        tangents_out,
        tree_util.tree_map(
            # compatible with JAX 0.4.34
            lambda a: (
                ad.Zero.from_primal_value(a)
                if jax.__version_info__ >= (0, 4, 34) else
                ad.Zero.from_value(a)
            ),
            val_out
        )
    )
    return val_out, r


def _add_tangents(xs, ys):
    """Element-wise addition of two tangent pytrees.

    Parameters
    ----------
    xs : pytree
        First tangent pytree (may contain ``ad.Zero`` leaves).
    ys : pytree
        Second tangent pytree (may contain ``ad.Zero`` leaves).

    Returns
    -------
    pytree
        The element-wise sum of *xs* and *ys*.
    """
    return tree_util.tree_map(ad.add_tangents, xs, ys, is_leaf=lambda a: isinstance(a, ad.Zero))


def general_batching_rule(prim, args, axes, **kwargs):
    """General-purpose batching rule for custom JAX primitives.

    Implements batching by separating batched and non-batched arguments,
    moving all batch dimensions to axis 0, and then applying the primitive
    to each element in the batch via ``jax.lax.scan``.

    This function is registered as the default batching rule for every
    :class:`XLACustomKernel` during initialization.

    Parameters
    ----------
    prim : Primitive
        The JAX primitive operation to be batched.
    args : sequence of array_like
        Input arguments to the primitive.
    axes : sequence of int or None
        Batch dimension index for each argument.  ``None`` indicates
        that the corresponding argument is not batched.
    **kwargs
        Additional keyword arguments forwarded to the primitive.

    Returns
    -------
    outs : pytree
        The batched outputs from applying the primitive.
    out_dim : pytree
        A pytree with the same structure as *outs*, where every leaf
        is ``0``, indicating that the batch dimension is the leading
        axis of each output.

    Notes
    -----
    All batch dimensions are moved to axis 0 before scanning.  The scan
    carry is unused (always ``0``); only the stacked scan outputs are
    returned.

    See Also
    --------
    XLACustomKernel.register_general_batching : Registers this function
        as the batching rule for a primitive.
    XLACustomKernel.def_batching_rule : Override with a custom batching
        rule.

    Examples
    --------
    .. code-block:: python

        >>> import functools
        >>> from jax.interpreters import batching
        >>> batching.primitive_batchers[my_prim] = functools.partial(
        ...     general_batching_rule, my_prim
        ... )
    """
    batch_axes, batch_args, non_batch_args = [], {}, {}
    for ax_i, ax in enumerate(axes):
        if ax is None:
            non_batch_args[f'ax{ax_i}'] = args[ax_i]
        else:
            batch_args[f'ax{ax_i}'] = args[ax_i] if ax == 0 else jax.numpy.moveaxis(args[ax_i], ax, 0)
            batch_axes.append(ax_i)

    def f(_, x):
        """
        Internal function for jax.lax.scan that applies the primitive to a single batch element.

        Args:
            _: Carry value (unused).
            x: Dictionary containing the current batch slice for each batched argument.

        Returns:
            tuple: (carry value, primitive output)
        """
        pars = tuple(
            [(x[f'ax{i}'] if i in batch_axes else non_batch_args[f'ax{i}'])
             for i in range(len(axes))]
        )
        return 0, prim(*pars, **kwargs)

    _, outs = jax.lax.scan(f, 0, batch_args)
    out_vals, out_tree = jax.tree.flatten(outs)
    out_dim = jax.tree.unflatten(out_tree, (0,) * len(out_vals))
    return outs, out_dim


class ShapeDtype(Protocol):
    """Protocol for objects that expose ``shape`` and ``dtype`` attributes.

    This protocol is used for type hinting to indicate that an object
    provides information about its tensor shape (as a tuple of integers)
    and its data type (as a NumPy dtype).  It is commonly satisfied by
    ``jax.ShapeDtypeStruct``, ``jax.core.ShapedArray``, and any
    user-defined class with the matching attributes.

    See Also
    --------
    OutType : Type alias combining ``ShapeDtype`` and sequences thereof.

    Examples
    --------
    .. code-block:: python

        >>> import numpy as np
        >>> from typing import Tuple
        >>>
        >>> class MyTensorSpec:
        ...     def __init__(self, shape: Tuple[int, ...], dtype: np.dtype):
        ...         self._shape = shape
        ...         self._dtype = dtype
        ...
        ...     @property
        ...     def shape(self) -> Tuple[int, ...]:
        ...         return self._shape
        ...
        ...     @property
        ...     def dtype(self) -> np.dtype:
        ...         return self._dtype
        >>>
        >>> def process_spec(spec: ShapeDtype):
        ...     print(f"Shape: {spec.shape}, Dtype: {spec.dtype}")
        >>>
        >>> spec = MyTensorSpec(shape=(10, 20), dtype=np.float32)
        >>> process_spec(spec)
        Shape: (10, 20), Dtype: float32
    """

    @property
    def shape(self) -> Tuple[int, ...]:
        """The shape of the tensor as a tuple of integers."""
        ...

    @property
    def dtype(self) -> np.dtype:
        """The data type of the tensor elements (e.g., ``np.float32``)."""
        ...


OutType = Union[ShapeDtype, Sequence[ShapeDtype]]


def _transform_to_shapedarray(a):
    """Convert an object with ``shape`` and ``dtype`` to a ``jax.core.ShapedArray``.

    Parameters
    ----------
    a : ShapeDtype
        Any object satisfying the :class:`ShapeDtype` protocol.

    Returns
    -------
    jax.core.ShapedArray
        The corresponding JAX abstract array.
    """
    return jax.core.ShapedArray(a.shape, a.dtype)


def abstract_arguments(outs):
    """Flatten an output specification into abstract ``ShapedArray`` objects.

    Takes a pytree of objects with ``shape`` and ``dtype`` attributes
    (e.g., ``jax.ShapeDtypeStruct``) and returns a flat list of
    ``jax.core.ShapedArray`` instances together with the tree definition
    needed to reconstruct the original structure.

    Parameters
    ----------
    outs : OutType
        An output specification -- a single ``ShapeDtype`` object or a
        pytree (list, tuple, dict, ...) of them.

    Returns
    -------
    flat_outs : list of jax.core.ShapedArray
        Flattened list of abstract arrays.
    tree_def : jax.tree_util.PyTreeDef
        The pytree definition that can be used to unflatten the outputs
        back into the original structure.

    See Also
    --------
    ShapeDtype : Protocol describing the expected input objects.

    Examples
    --------
    .. code-block:: python

        >>> import jax
        >>> import jax.numpy as jnp
        >>> spec = [
        ...     jax.ShapeDtypeStruct((10,), jnp.float32),
        ...     jax.ShapeDtypeStruct((5, 3), jnp.int32),
        ... ]
        >>> flat, treedef = abstract_arguments(spec)
        >>> len(flat)
        2
        >>> flat[0].shape
        (10,)
    """
    outs = jax.tree.map(_transform_to_shapedarray, outs)
    outs, tree_def = jax.tree.flatten(outs)
    return outs, tree_def


def jaxtype_to_warptype(dtype):
    """Convert a JAX / NumPy dtype to the corresponding Warp scalar type.

    Maps standard NumPy data types (which are also used by JAX) to their
    Warp equivalents.  This is needed when constructing Warp kernel
    signatures or Warp array types from JAX metadata.

    Parameters
    ----------
    dtype : numpy.dtype or type
        The data type to convert.  Accepts any object that can be
        compared with NumPy scalar types (e.g., ``np.float32``,
        ``jnp.float32``, ``np.dtype('float32')``).

    Returns
    -------
    warp type
        The corresponding Warp scalar type (e.g., ``warp.float32``,
        ``warp.int32``, ``warp.bool``).

    Raises
    ------
    ImportError
        If the ``warp`` package is not installed.
    ValueError
        If *dtype* does not correspond to any supported Warp type.
        Supported types include: ``float16``, ``float32``, ``float64``,
        ``int8``, ``int16``, ``int32``, ``int64``, ``uint8``,
        ``uint16``, ``uint32``, ``uint64``, and ``bool_``.

    See Also
    --------
    jaxinfo_to_warpinfo : Convert a full ``jax.ShapeDtypeStruct`` to a
        Warp array type.
    check_warp_installed : Verify that Warp is available.

    Notes
    -----
    The mapping covers all scalar types supported by both NumPy and
    Warp: ``float16``, ``float32``, ``float64``, ``int8`` through
    ``int64``, ``uint8`` through ``uint64``, and ``bool_``.  Complex
    types are not supported by Warp and will raise ``ValueError``.

    Examples
    --------
    .. code-block:: python

        >>> import numpy as np
        >>> warp_type = jaxtype_to_warptype(np.float32)
        >>> warp_type  # warp.float32
    """
    if not warp_installed:
        raise ImportError('Warp is required to convert JAX dtypes to Warp types.')

    # float
    if dtype == np.float16:
        return warp.float16
    elif dtype == np.float32:
        return warp.float32
    elif dtype == np.float64:
        return warp.float64

    # integer
    elif dtype == np.int8:
        return warp.int8
    elif dtype == np.int16:
        return warp.int16
    elif dtype == np.int32:
        return warp.int32
    elif dtype == np.int64:
        return warp.int64

    # unsigned integer
    elif dtype == np.uint8:
        return warp.uint8
    elif dtype == np.uint16:
        return warp.uint16
    elif dtype == np.uint32:
        return warp.uint32
    elif dtype == np.uint64:
        return warp.uint64

    # boolean
    elif dtype == np.bool_:
        return warp.bool
    else:
        raise ValueError(f"Warp does not support computations with dtype: {dtype}")


def jaxinfo_to_warpinfo(jax_info: jax.ShapeDtypeStruct):
    """Convert a ``jax.ShapeDtypeStruct`` to a Warp array type descriptor.

    Takes a JAX shape-and-dtype specification and creates the
    corresponding Warp array type with matching data type and
    dimensionality.  This is useful when building Warp kernel
    signatures from JAX output specifications.

    Parameters
    ----------
    jax_info : jax.ShapeDtypeStruct
        A JAX structure containing ``shape``, ``dtype``, and ``ndim``
        attributes describing an array.

    Returns
    -------
    warp.types.array
        A Warp array type with matching data type and number of
        dimensions, suitable for use in Warp kernel definitions.

    Raises
    ------
    ImportError
        If the ``warp`` package is not installed (propagated from
        :func:`jaxtype_to_warptype`).
    ValueError
        If the dtype in *jax_info* is not supported by Warp (propagated
        from :func:`jaxtype_to_warptype`).

    See Also
    --------
    jaxtype_to_warptype : Convert a single dtype to a Warp type.
    check_warp_installed : Verify that Warp is available.

    Notes
    -----
    The resulting Warp array type is constructed via
    ``warp.array(dtype=..., ndim=...)`` which creates a Warp type
    descriptor (not an actual array).  This is typically used in Warp
    kernel function signatures to define input/output types.

    Examples
    --------
    .. code-block:: python

        >>> import jax
        >>> import numpy as np
        >>> info = jax.ShapeDtypeStruct(shape=(32, 32), dtype=np.float32)
        >>> warp_arr_type = jaxinfo_to_warpinfo(info)
    """
    dtype = jaxtype_to_warptype(jax_info.dtype)
    return warp.array(dtype=dtype, ndim=jax_info.ndim)



# Copyright 2026 BrainX Ecosystem Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from typing import Optional

import brainunit as u
import jax
import jax.numpy as jnp
import numpy as np

from brainevent._misc import generate_block_dim, namescope
from brainevent._op import numba_kernel, jaxinfo_to_warpinfo, XLACustomKernel
from brainevent._op.benchmark import BenchmarkConfig
from brainevent._typing import Data, Indptr, Index, MatrixShape

__all__ = [
    'csrmv_yw2y',
    'csrmv_yw2y_p',
]


@namescope(static_argnames=['shape', 'transpose'])
def csrmv_yw2y(
    y: Data,
    w: Data,
    indices: Index,
    indptr: Indptr,
    *,
    shape,
    transpose: bool = False,
    backend: Optional[str] = None,
) -> Data:
    """
    Element-wise product of a vector and CSR weights, indexed by CSR
    structure.

    For each non-zero entry ``j`` in the CSR matrix at position
    ``(row, col)``, computes ``out[j] = w[j] * y[row]`` (non-transposed)
    or ``out[j] = w[j] * y[col]`` (transposed).  The result has the same
    shape as ``w`` and ``indices`` (i.e., one value per structural
    non-zero).

    This operation is useful for computing per-synapse quantities in
    neural network models where ``y`` is a neuron-level vector and ``w``
    contains per-synapse weights stored in CSR form.

    The function supports physical units via :mod:`brainunit`.

    Parameters
    ----------
    y : jax.Array, numpy.ndarray, or brainunit.Quantity
        Dense vector indexed by the CSR structure.  Shape
        ``(shape[0],)`` when ``transpose=False`` or ``(shape[1],)`` when
        ``transpose=True``.
    w : jax.Array, numpy.ndarray, or brainunit.Quantity
        Per-synapse weight values.  Shape ``(nse,)``, must match the
        shape of ``indices``.
    indices : jax.Array or numpy.ndarray
        Column indices of the CSR matrix.  Shape ``(nse,)`` with integer
        dtype.
    indptr : jax.Array or numpy.ndarray
        Row index pointer array.  Shape ``(shape[0] + 1,)`` with integer
        dtype.
    shape : tuple of int
        Two-element tuple ``(m, k)`` giving the logical shape of the
        CSR matrix.
    transpose : bool, optional
        If ``True``, index ``y`` by column indices instead of row indices.
        Default is ``False``.
    backend : str or None, optional
        Compute backend.  Default is ``None`` (auto-select).

    Returns
    -------
    out : jax.Array or brainunit.Quantity
        Per-synapse result vector.  Shape ``(nse,)``, same as ``w``.

    See Also
    --------
    csrmv : Standard CSR matrix--vector multiplication.

    Notes
    -----
    This operation is differentiable with respect to both ``y`` and ``w``
    via custom JVP rules.  The transpose rule is not yet implemented.

    Mathematically, for each structural non-zero entry ``j`` of the CSR
    matrix at position ``(row, col)``, the output is computed as:

    ``out[j] = w[j] * y[row]``  (non-transposed, ``transpose=False``)

    ``out[j] = w[j] * y[col]``  (transposed, ``transpose=True``)

    where ``row`` is determined by the ``indptr`` array (the row to
    which the ``j``-th non-zero belongs) and ``col = indices[j]``.

    This operation is distinct from standard sparse matrix--vector
    multiplication: it produces one output element per structural
    non-zero rather than one per matrix row.  It is commonly used to
    compute per-synapse quantities in spiking neural network models.

    Examples
    --------
    .. code-block:: python

        >>> import jax.numpy as jnp
        >>> from brainevent._csr.float import csrmv_yw2y
        >>> y = jnp.array([1.0, 2.0])
        >>> w = jnp.array([0.5, 0.3, 0.7, 0.1])
        >>> indices = jnp.array([0, 2, 1, 2], dtype=jnp.int32)
        >>> indptr = jnp.array([0, 2, 4], dtype=jnp.int32)
        >>> csrmv_yw2y(y, w, indices, indptr, shape=(2, 3))
    """
    w, w_unit = u.split_mantissa_unit(w)
    y, _ = u.split_mantissa_unit(y)
    res = csrmv_yw2y_p_call(y, w, indices, indptr, shape=shape, transpose=transpose, backend=backend)[0]
    return u.maybe_decimal(res * w_unit)


def _csrmv_yw2y_numba_kernels(
    transpose: bool,
    **kwargs
):
    import numba  # pylint: disable=import-outside-toplevel

    if transpose:
        @numba.njit
        def mm(y, w, indices, indptr, posts):
            for i_col in range(indptr.shape[0] - 1):
                i_row_start = indptr[i_col]
                i_row_end = indptr[i_col + 1]
                for j in range(i_row_start, i_row_end):
                    posts[j] = w[j] * y[indices[j]]

        def kernel(y, w, indices, indptr):
            return numba_kernel(mm, outs=kwargs['outs'])(y, w, indices, indptr)

    else:
        @numba.njit
        def mm(y, w, indptr, posts):
            for i_row in range(indptr.shape[0] - 1):
                i_col_start = indptr[i_row]
                i_col_end = indptr[i_row + 1]
                for j in range(i_col_start, i_col_end):
                    posts[j] = w[j] * y[i_row]

        def kernel(y, w, indices, indptr):
            return numba_kernel(mm, outs=kwargs['outs'])(y, w, indptr)

    return kernel


def _csrmv_yw2y_pallas_kernels(
    shape: MatrixShape,
    transpose: bool,
    y_info: jax.ShapeDtypeStruct,
    **kwargs
):
    from jax.experimental import pallas as pl
    from jax.experimental.pallas.triton import load, store

    block_dim = generate_block_dim(y_info.shape[0], 128)

    if transpose:
        def mm(
            y_ref,
            w_ref,
            indices_ref,
            indptr_ref,
            posts_ref,
        ):
            i_block = pl.program_id(0)
            num_blocks_grid = indptr_ref.shape[0] - 1

            def _body():
                i_start = indptr_ref[i_block]
                i_end = indptr_ref[i_block + 1]
                num_blocks = (i_end - i_start + block_dim - 1) // block_dim

                def loop_fn(i, _):
                    offset = i_start + i * block_dim
                    mask = (offset + jnp.arange(block_dim)) < i_end

                    w = load(w_ref.at[pl.ds(offset, block_dim)], mask=mask, other=0.0)
                    index = load(indices_ref.at[pl.ds(offset, block_dim)], mask=mask, other=0)

                    # Indirect Read Guard
                    valid_idx = index < y_ref.shape[0]
                    safe_index = jnp.minimum(index, y_ref.shape[0] - 1)

                    # Combine Masks
                    final_mask = mask & valid_idx

                    y = y_ref[safe_index]
                    y = jnp.where(final_mask, y, 0.0)

                    store(posts_ref.at[pl.ds(offset, block_dim)], w * y, mask=final_mask)

                jax.lax.fori_loop(0, num_blocks, loop_fn, None)

            # GUARD 1: Grid / Indptr Boundary Check using jax.lax.cond
            jax.lax.cond(i_block < num_blocks_grid, _body, lambda: None)

        def kernel(y, w, indices, indptr):
            fn = pl.pallas_call(
                mm,
                grid=(shape[0],),
                out_shape=kwargs['outs'],
                backend='triton'
            )
            return fn(y, w, indices, indptr)
    else:
        def mm(
            y_ref,
            w_ref,
            indptr_ref,
            posts_ref,
        ):
            i_block = pl.program_id(0)
            num_blocks_grid = indptr_ref.shape[0] - 1

            def _body():
                i_start = indptr_ref[i_block]
                i_end = indptr_ref[i_block + 1]
                num_blocks = (i_end - i_start + block_dim - 1) // block_dim
                y_scalar = y_ref[i_block]

                def loop_fn(i, _):
                    offset = i_start + i * block_dim
                    mask = (offset + jnp.arange(block_dim)) < i_end

                    w = load(w_ref.at[pl.ds(offset, block_dim)], mask=mask, other=0.0)
                    store(posts_ref.at[pl.ds(offset, block_dim)], w * y_scalar, mask=mask)

                jax.lax.fori_loop(0, num_blocks, loop_fn, None)

            # GUARD 1: Grid / Indptr Boundary Check using jax.lax.cond
            jax.lax.cond(i_block < num_blocks_grid, _body, lambda: None)

        def kernel(y, w, indices, indptr):
            fn = pl.pallas_call(mm, grid=(shape[0],), out_shape=kwargs['outs'], backend='triton')
            return fn(y, w, indptr)

    return kernel


def _csrmv_yw2y_warp_kernel_generator(
    y_info: jax.ShapeDtypeStruct,
    w_info: jax.ShapeDtypeStruct,
    indices_info: jax.ShapeDtypeStruct,
    indptr_info: jax.ShapeDtypeStruct,
    transpose: bool,
    shape: MatrixShape,
    **kwargs
):
    import warp  # pylint: disable=import-outside-toplevel
    from warp.jax_experimental import jax_kernel

    y_warp_info = jaxinfo_to_warpinfo(y_info)
    w_warp_info = jaxinfo_to_warpinfo(w_info)
    indices_warp_info = jaxinfo_to_warpinfo(indices_info)
    indptr_warp_info = jaxinfo_to_warpinfo(indptr_info)
    out_warp_info = jaxinfo_to_warpinfo(kwargs['outs'][0])

    if transpose:
        @warp.kernel
        def yw2y(
            y: y_warp_info,
            w: w_warp_info,
            indices: indices_warp_info,
            indptr: indptr_warp_info,
            posts: out_warp_info,
        ):
            i_row = warp.tid()
            for j in range(indptr[i_row], indptr[i_row + 1]):
                posts[j] = w[j] * y[indices[j]]

        def kernel(y, w, indices, indptr):
            out_info = kwargs['outs'][0]
            dim = shape[0]
            fn = jax_kernel(yw2y, launch_dims=[dim], num_outputs=1, output_dims={'posts': out_info.shape})
            return fn(y, w, indices, indptr)

    else:
        @warp.kernel
        def yw2y(
            y: y_warp_info,
            w: w_warp_info,
            indptr: indptr_warp_info,
            posts: out_warp_info,
        ):
            i_row = warp.tid()
            y_val = y[i_row]
            for j in range(indptr[i_row], indptr[i_row + 1]):
                posts[j] = w[j] * y_val

        def kernel(y, w, indices, indptr):
            out_info = kwargs['outs'][0]
            dim = shape[0]
            fn = jax_kernel(yw2y, launch_dims=[dim], num_outputs=1, output_dims={'posts': out_info.shape})
            return fn(y, w, indptr)

    return kernel


def _csrmv_yw2y_jvp_y(y_dot, y, w, indices, indptr, *, shape, transpose, **kwargs):
    return csrmv_yw2y_p_call(y_dot, w, indices, indptr, shape=shape, transpose=transpose, backend=kwargs['backend'])


def _csrmv_yw2y_jvp_w(w_dot, y, w, indices, indptr, *, shape, transpose, **kwargs):
    return csrmv_yw2y_p_call(y, w_dot, indices, indptr, shape=shape, transpose=transpose, backend=kwargs['backend'])


def _csrmv_yw2y_transpose_rule(ct, y, w, indices, indptr, *, shape, transpose, **kwargs):
    raise NotImplementedError


def _csrmv_yw2y_benchmark_data(*, platform):
    n_pre, n_post, prob, dtype = 1000, 1000, 0.1, jnp.float32
    configs = []
    for transpose in (False, True):
        n_conn = max(1, int(n_post * prob))
        indptr = np.arange(n_pre + 1, dtype=np.int32) * n_conn
        indices = np.random.randint(0, n_post, (n_pre * n_conn,), dtype=np.int32)
        w = jnp.ones(n_pre * n_conn, dtype=dtype)
        y_size = n_pre if not transpose else n_post
        y = jnp.asarray(np.random.randn(y_size), dtype=dtype)
        name = f"{'T' if transpose else 'NT'}"
        configs.append(
            BenchmarkConfig(
                name,
                (y, w, indices, jnp.asarray(indptr)),
                {'shape': (n_pre, n_post), 'transpose': transpose}
            )
        )
    return configs


def csrmv_yw2y_p_call(
    y: Data,
    w: Data,
    indices: Index,
    indptr: Indptr,
    *,
    shape: MatrixShape,
    transpose: bool = False,
    backend: Optional[str] = None,
):
    """
    Low-level primitive call for the element-wise vector--weight product
    indexed by CSR structure.

    Validates inputs and dispatches the ``csrmv_yw2y_p`` XLA custom kernel.
    For each structural non-zero ``j`` at position ``(row, col)`` in the
    CSR matrix, the output is ``out[j] = w[j] * y[row]`` (non-transposed)
    or ``out[j] = w[j] * y[col]`` (transposed).

    Parameters
    ----------
    y : jax.Array
        Dense vector.  Shape ``(shape[0],)`` when ``transpose=False`` or
        ``(shape[1],)`` when ``transpose=True``.
    w : jax.Array
        Per-synapse weight values.  Shape ``(nse,)``, must match the shape
        of ``indices``.
    indices : jax.Array
        Column indices of the CSR matrix.  Shape ``(nse,)`` with integer
        dtype.
    indptr : jax.Array
        Row index pointer array.  Shape ``(shape[0] + 1,)`` with integer
        dtype.
    shape : tuple of int
        Two-element tuple ``(m, k)`` giving the logical shape of the CSR
        matrix.
    transpose : bool, optional
        If ``True``, index ``y`` by column indices.  Default is ``False``.
    backend : str or None, optional
        Compute backend to use.  Default is ``None`` (auto-select).

    Returns
    -------
    list of jax.Array
        A single-element list containing the per-synapse result.  Shape
        ``(nse,)``, same as ``w``.

    Raises
    ------
    AssertionError
        If ``y`` and ``w`` have different dtypes.
    AssertionError
        If ``y`` or ``w`` is not 1-D.
    AssertionError
        If ``indices`` or ``indptr`` does not have an integer dtype.
    AssertionError
        If ``w`` does not have a floating-point dtype.
    AssertionError
        If ``w`` and ``indices`` do not have the same shape.
    AssertionError
        If there is a shape mismatch between ``y`` and the sparse
        matrix ``shape`` (considering the ``transpose`` flag).

    See Also
    --------
    csrmv_yw2y : High-level wrapper with unit support.

    Notes
    -----
    The computation performed is, for each structural non-zero entry
    ``j`` at CSR position ``(row, col)``:

    ``out[j] = w[j] * y[row]``  (non-transposed)

    ``out[j] = w[j] * y[col]``  (transposed)

    where ``row`` is determined by ``indptr`` and ``col = indices[j]``.

    Examples
    --------
    .. code-block:: python

        >>> import jax.numpy as jnp
        >>> from brainevent import csrmv_yw2y
        >>> y = jnp.array([1.0, 2.0])
        >>> w = jnp.array([0.5, 0.3, 0.7, 0.1])
        >>> indices = jnp.array([0, 2, 1, 2], dtype=jnp.int32)
        >>> indptr = jnp.array([0, 2, 4], dtype=jnp.int32)
        >>> result = csrmv_yw2y(
        ...     y, w, indices, indptr,
        ...     shape=(2, 3), transpose=False)
    """
    assert y.dtype == w.dtype, f"y and w must have the same dtype, but got {y.dtype} and {w.dtype}."
    assert indptr.ndim == 1, "Indptr must be 1D."
    assert indices.ndim == 1, "Indices must be 1D."
    assert y.ndim == w.ndim == 1, "y and w must have the same shape."
    assert jnp.issubdtype(indices.dtype, jnp.integer), "Indices must be an integer type."
    assert jnp.issubdtype(indptr.dtype, jnp.integer), "indptr must be an integer type."
    assert jnp.issubdtype(w.dtype, jnp.floating), 'Weights must be a floating-point type.'
    assert w.shape == indices.shape, f"Weights shape mismatch, expected {indices.shape}, got {w.shape}."
    if transpose:
        # [x] @ [h, w] -> [w]
        assert shape[1] == y.shape[0], "Shape mismatch for transpose operation."
    else:
        # [h, w] @ [x] -> [h]
        assert shape[0] == y.shape[0], "Shape mismatch for non-transpose operation."

    return csrmv_yw2y_p(
        y,
        w,
        indices,
        indptr,
        outs=[jax.ShapeDtypeStruct(w.shape, w.dtype)],
        shape=tuple(shape),
        transpose=transpose,
        backend=backend,
        indices_info=jax.ShapeDtypeStruct(indices.shape, indices.dtype),
        indptr_info=jax.ShapeDtypeStruct(indptr.shape, indptr.dtype),
        y_info=jax.ShapeDtypeStruct(y.shape, y.dtype),
        w_info=jax.ShapeDtypeStruct(w.shape, w.dtype),
    )


csrmv_yw2y_p = XLACustomKernel(
    'csrmv_yw2y',
    doc="""
Low-level XLA custom-kernel primitive for ``csrmv_yw2y``.

This ``XLACustomKernel`` instance dispatches the CSR sparse matrix-vector multiplication with element-wise product (y * W -> y)
operation to registered backends (``numba``, ``warp``, ``pallas``),
using runtime shape/dtype metadata provided by the high-level wrapper.

For each non-zero entry at position (row, col) in the CSR matrix, computes
out[j] = w[j] * y[row] (non-transposed) or out[j] = w[j] * y[col] (transposed),
producing one output element per structural non-zero.

Beyond backend dispatch, the primitive stores JAX transformation bindings
(JVP, transpose, batching, and call registration) so the operation integrates
correctly with ``jit``, ``vmap``, and autodiff.

Available backends can be queried with ``csrmv_yw2y_p.available_backends(platform)``,
and the default backend can be configured with ``csrmv_yw2y_p.set_default(platform, backend)``.

See Also
--------
csrmv_yw2y : High-level user-facing function wrapper.
"""
)
csrmv_yw2y_p.def_numba_kernel(_csrmv_yw2y_numba_kernels)
csrmv_yw2y_p.def_warp_kernel(_csrmv_yw2y_warp_kernel_generator)
csrmv_yw2y_p.def_pallas_kernel('gpu', _csrmv_yw2y_pallas_kernels)
csrmv_yw2y_p.def_jvp_rule2(_csrmv_yw2y_jvp_y, _csrmv_yw2y_jvp_w, None, None)
csrmv_yw2y_p.def_call(csrmv_yw2y_p_call)
csrmv_yw2y_p.def_tags('csr', 'float')
csrmv_yw2y_p.def_benchmark_data(_csrmv_yw2y_benchmark_data)



# Copyright 2024 BrainX Ecosystem Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import brainstate
import jax
import jax.numpy as jnp
import numpy as np
import pytest

import brainevent
from brainevent._test_util import allclose, gen_events

platform = jax.default_backend()

if platform == 'cpu':
    shapes = [
        (200, 300),
        (100, 500)
    ]
else:
    shapes = [
        (2000, 3000),
        (1000, 5000)
    ]


class Test_JITC_RC_Conversion:

    @pytest.mark.parametrize('shape', shapes)
    @pytest.mark.parametrize('corder', [True, False])
    def test_matvec(self, shape, corder):
        jitcr = brainevent.JITCScalarR((1.5, 0.1, 123), shape=shape, corder=corder)
        jitcc = jitcr.T

        vector = jnp.asarray(np.random.rand(shape[1]))

        out1 = jitcr @ vector
        out2 = vector @ jitcc
        assert allclose(out1, out2)
        jax.block_until_ready((vector, out1, out2))

    @pytest.mark.parametrize('shape', shapes)
    @pytest.mark.parametrize('corder', [True, False])
    def test_vecmat(self, shape, corder):
        jitcr = brainevent.JITCScalarR((1.5, 0.1, 123), shape=shape, corder=corder)
        jitcc = jitcr.T

        vector = jnp.asarray(np.random.rand(shape[0]))

        out1 = vector @ jitcr
        out2 = jitcc @ vector
        assert allclose(out1, out2)
        jax.block_until_ready((vector, out1, out2))

    @pytest.mark.parametrize('k', [10])
    @pytest.mark.parametrize('shape', shapes)
    @pytest.mark.parametrize('corder', [True, False])
    def test_jitmat(self, k, shape, corder):
        jitcr = brainevent.JITCScalarR((1.5, 0.1, 123), shape=shape, corder=corder)
        jitcc = jitcr.T

        matrix = jnp.asarray(np.random.rand(shape[1], k))

        out1 = jitcr @ matrix
        out2 = (matrix.T @ jitcc).T
        assert allclose(out1, out2, rtol=1e-3, atol=1e-3)
        jax.block_until_ready((matrix, out1, out2))

    @pytest.mark.parametrize('k', [10])
    @pytest.mark.parametrize('shape', shapes)
    @pytest.mark.parametrize('corder', [True, False])
    def test_matjit(self, k, shape, corder):
        jitcr = brainevent.JITCScalarR((1.5, 0.1, 123), shape=shape, corder=corder)
        jitcc = jitcr.T

        matrix = jnp.asarray(np.random.rand(k, shape[0]))

        out1 = matrix @ jitcr
        out2 = (jitcc @ matrix.T).T
        assert allclose(out1, out2, atol=1e-4, rtol=1e-4)
        jax.block_until_ready((matrix, out1, out2))

    @pytest.mark.parametrize('shape', shapes)
    @pytest.mark.parametrize('corder', [True, False])
    @pytest.mark.parametrize('asbool', [True, False])
    def test_matvec_event(self, shape, corder, asbool):
        jitcr = brainevent.JITCScalarR((1.5, 0.1, 123), shape=shape, corder=corder)
        jitcc = jitcr.T

        vector = gen_events(shape[1], asbool=asbool)

        out1 = jitcr @ vector
        out2 = vector @ jitcc
        assert allclose(out1, out2)
        jax.block_until_ready((out1, out2))

    @pytest.mark.parametrize('shape', shapes)
    @pytest.mark.parametrize('corder', [True, False])
    @pytest.mark.parametrize('asbool', [True, False])
    def test_vecmat_event(self, shape, corder, asbool):
        jitcr = brainevent.JITCScalarR((1.5, 0.1, 123), shape=shape, corder=corder)
        jitcc = jitcr.T

        vector = gen_events(shape[0], asbool=asbool)

        out1 = vector @ jitcr
        out2 = jitcc @ vector
        assert allclose(out1, out2)
        jax.block_until_ready((out1, out2))

    @pytest.mark.parametrize('k', [10])
    @pytest.mark.parametrize('shape', shapes)
    @pytest.mark.parametrize('corder', [True, False])
    @pytest.mark.parametrize('asbool', [True, False])
    def test_jitmat_event(self, k, shape, corder, asbool):
        jitcr = brainevent.JITCScalarR((1.5, 0.1, 123), shape=shape, corder=corder)
        jitcc = jitcr.T

        matrix = gen_events([shape[1], k], asbool=asbool)

        out1 = jitcr @ matrix
        out2 = (matrix.value.T @ jitcc).T
        assert allclose(out1, out2)
        jax.block_until_ready((out1, out2))

    @pytest.mark.parametrize('k', [10])
    @pytest.mark.parametrize('shape', shapes)
    @pytest.mark.parametrize('corder', [True, False])
    @pytest.mark.parametrize('asbool', [True, False])
    def test_matjit_event(self, k, shape, corder, asbool):
        jitcr = brainevent.JITCScalarR((1.5, 0.1, 123), shape=shape, corder=corder)
        jitcc = jitcr.T

        matrix = gen_events([k, shape[0]], asbool=asbool)

        out1 = matrix @ jitcr
        out2 = (jitcc @ matrix.value.T).T
        print(out1 - out2)
        assert allclose(out1, out2, atol=1e-4, rtol=1e-4)
        jax.block_until_ready((out1, out2))


class Test_JITC_Operator_Behavior:
    @pytest.mark.parametrize('corder', [True, False])
    def test_jitc_scalar_r_operator_behavior(self, corder):
        shape = (20, 30)
        mat = brainevent.JITCScalarR((1.5, 0.1, 123), shape=shape, corder=corder)
        dense = mat.todense()

        left_vec = gen_events(shape[0], asbool=False).value
        right_vec = gen_events(shape[1], asbool=False).value
        left_mat = gen_events((5, shape[0]), asbool=False).value
        right_mat = gen_events((shape[1], 4), asbool=False).value

        r1 = left_vec @ mat
        r2 = left_vec @ dense
        r3 = mat @ right_vec
        r4 = dense @ right_vec
        r5 = left_mat @ mat
        r6 = left_mat @ dense
        r7 = mat @ right_mat
        r8 = dense @ right_mat
        assert allclose(r1, r2, atol=1e-4, rtol=1e-4)
        assert allclose(r3, r4, atol=1e-4, rtol=1e-4)
        assert allclose(r5, r6, atol=1e-4, rtol=1e-4)
        assert allclose(r7, r8, atol=1e-4, rtol=1e-4)
        jax.block_until_ready((dense, left_vec, right_vec, left_mat, right_mat, r1, r2, r3, r4, r5, r6, r7, r8))

    @pytest.mark.parametrize('corder', [True, False])
    def test_jitc_scalar_c_operator_behavior(self, corder):
        shape = (20, 30)
        mat = brainevent.JITCScalarC((1.5, 0.1, 123), shape=shape, corder=corder)
        dense = mat.todense()

        left_vec = gen_events(shape[0], asbool=False).value
        right_vec = gen_events(shape[1], asbool=False).value
        left_mat = gen_events((5, shape[0]), asbool=False).value
        right_mat = gen_events((shape[1], 4), asbool=False).value

        r1 = left_vec @ mat
        r2 = left_vec @ dense
        r3 = mat @ right_vec
        r4 = dense @ right_vec
        r5 = left_mat @ mat
        r6 = left_mat @ dense
        r7 = mat @ right_mat
        r8 = dense @ right_mat
        assert allclose(r1, r2, atol=1e-4, rtol=1e-4)
        assert allclose(r3, r4, atol=1e-4, rtol=1e-4)
        assert allclose(r5, r6, atol=1e-4, rtol=1e-4)
        assert allclose(r7, r8, atol=1e-4, rtol=1e-4)
        jax.block_until_ready((dense, left_vec, right_vec, left_mat, right_mat, r1, r2, r3, r4, r5, r6, r7, r8))

    @pytest.mark.parametrize('corder', [True, False])
    def test_jitc_scalar_r_transpose_operator_behavior(self, corder):
        shape = (20, 30)
        mat = brainevent.JITCScalarR((1.5, 0.1, 123), shape=shape, corder=corder).T
        dense = mat.todense()

        left_vec = jnp.asarray(np.random.rand(shape[1]))
        right_vec = jnp.asarray(np.random.rand(shape[0]))
        left_mat = jnp.asarray(np.random.rand(5, shape[1]))
        right_mat = jnp.asarray(np.random.rand(shape[0], 4))

        r1 = left_vec @ mat
        r2 = left_vec @ dense
        r3 = mat @ right_vec
        r4 = dense @ right_vec
        r5 = left_mat @ mat
        r6 = left_mat @ dense
        r7 = mat @ right_mat
        r8 = dense @ right_mat
        assert allclose(r1, r2, atol=1e-4, rtol=1e-4)
        assert allclose(r3, r4, atol=1e-4, rtol=1e-4)
        assert allclose(r5, r6, atol=1e-4, rtol=1e-4)
        assert allclose(r7, r8, atol=1e-4, rtol=1e-4)
        jax.block_until_ready((dense, left_vec, right_vec, left_mat, right_mat, r1, r2, r3, r4, r5, r6, r7, r8))

    @pytest.mark.parametrize('corder', [True, False])
    def test_jitc_scalar_c_transpose_operator_behavior(self, corder):
        shape = (20, 30)
        mat = brainevent.JITCScalarC((1.5, 0.1, 123), shape=shape, corder=corder).T
        dense = mat.todense()

        left_vec = jnp.asarray(np.random.rand(shape[1]))
        right_vec = jnp.asarray(np.random.rand(shape[0]))
        left_mat = jnp.asarray(np.random.rand(5, shape[1]))
        right_mat = jnp.asarray(np.random.rand(shape[0], 4))

        r1 = left_vec @ mat
        r2 = left_vec @ dense
        r3 = mat @ right_vec
        r4 = dense @ right_vec
        r5 = left_mat @ mat
        r6 = left_mat @ dense
        r7 = mat @ right_mat
        r8 = dense @ right_mat
        assert allclose(r1, r2, atol=1e-4, rtol=1e-4)
        assert allclose(r3, r4, atol=1e-4, rtol=1e-4)
        assert allclose(r5, r6, atol=1e-4, rtol=1e-4)
        assert allclose(r7, r8, atol=1e-4, rtol=1e-4)
        jax.block_until_ready((dense, left_vec, right_vec, left_mat, right_mat, r1, r2, r3, r4, r5, r6, r7, r8))

    @pytest.mark.parametrize('cls', [brainevent.JITCScalarR, brainevent.JITCScalarC])
    def test_jitc_scalar_unit_operator_behavior(self, cls):
        import brainunit as u
        shape = (20, 30)
        weight = 2.1 * u.mV
        mat = cls((weight, 0.2, 123), shape=shape)
        dense = mat.todense()

        right_vec = jnp.asarray(np.random.rand(shape[1]))
        left_vec = jnp.asarray(np.random.rand(shape[0]))

        r1 = mat @ right_vec
        r2 = dense @ right_vec
        r3 = left_vec @ mat
        r4 = left_vec @ dense
        assert u.math.allclose(r1, r2,
                               rtol=1e-4 * u.get_unit(r2),
                               atol=1e-4 * u.get_unit(r2))
        assert u.math.allclose(r3, r4,
                               rtol=1e-4 * u.get_unit(r4),
                               atol=1e-4 * u.get_unit(r4))
        jax.block_until_ready((right_vec, left_vec))


class Test_JITC_To_Dense:

    @pytest.mark.parametrize('shape', shapes)
    @pytest.mark.parametrize('transpose', [True, False])
    @pytest.mark.parametrize('corder', [True, False])
    def test_todense(self, shape, transpose, corder):
        jitcr = brainevent.JITCScalarR((1.5, 0.1, 123), shape=shape, corder=corder)
        jitcc = jitcr.T

        out1 = jitcr.todense()
        out2 = jitcc.todense().T
        out3 = jitcr.T.todense().T
        out4 = jitcc.T.todense()
        assert allclose(out1, out2)
        assert allclose(out1, out3)
        assert allclose(out1, out4)
        jax.block_until_ready((out1, out2, out3, out4))

    @pytest.mark.parametrize('shape', shapes)
    @pytest.mark.parametrize('corder', [True, False])
    @pytest.mark.parametrize('weight', [-1., 1.])
    def test_vjp(self, shape, corder, weight):
        base = brainevent.JITCScalarR((1.0, 0.1, 123), shape=shape, corder=corder).todense()

        def f_dense_vjp(weight):
            res = base * weight
            return res

        ct = brainstate.random.random(shape)
        primals, f_vjp = jax.vjp(f_dense_vjp, weight)
        true_weight_grad, = f_vjp(ct)

        expected_weight_grad = (ct * base).sum()
        assert allclose(true_weight_grad, expected_weight_grad)

        def f_jitc_vjp(weight):
            mat = brainevent.JITCScalarR((weight, 0.1, 123), shape=shape, corder=corder)
            return mat.todense()

        primals, f_vjp2 = jax.vjp(f_jitc_vjp, weight)
        jitc_weight_grad, = f_vjp2(ct)

        assert allclose(true_weight_grad, jitc_weight_grad)
        jax.block_until_ready((base, ct, primals, true_weight_grad, expected_weight_grad, jitc_weight_grad))

    @pytest.mark.parametrize('shape', shapes)
    @pytest.mark.parametrize('corder', [True, False])
    @pytest.mark.parametrize('weight', [-1., 1.])
    def test_jvp(self, shape, corder, weight):
        base = brainevent.JITCScalarR((1., 0.1, 123), shape=shape, corder=corder).todense()
        tagents = (brainstate.random.random(),)

        def f_dense_jvp(weight):
            res = base * weight
            return res

        def f_jitc_jvp(weight):
            mat = brainevent.JITCScalarR((weight, 0.1, 123), shape=shape, corder=corder)
            return mat.todense()

        primals, true_grad = jax.jvp(f_dense_jvp, (weight,), tagents)
        primals, jitc_grad = jax.jvp(f_jitc_jvp, (weight,), tagents)
        assert allclose(true_grad, jitc_grad)
        jax.block_until_ready((base, tagents[0], primals, true_grad, jitc_grad))


class Test_JITC_Scalar_Validation:
    @pytest.mark.parametrize('cls', [brainevent.JITCScalarR, brainevent.JITCScalarC])
    @pytest.mark.parametrize('prob', [-0.1, 1.1, float('nan')])
    def test_invalid_prob_raises(self, cls, prob):
        with pytest.raises(ValueError, match='prob'):
            cls((1.5, prob, 123), shape=(8, 6))

    # @pytest.mark.parametrize('cls', [brainevent.JITCScalarR, brainevent.JITCScalarC])
    # @pytest.mark.parametrize('corder', [True, False])
    # def test_zero_prob_dense_matvec_matmat(self, cls, corder):
    #     shape = (8, 6)
    #     mat = cls((1.5, 0.0, 123), shape=shape, corder=corder)
    #
    #     dense = mat.todense()
    #     assert allclose(dense, jnp.zeros_like(dense))
    #
    #     vec = jnp.ones(shape[1])
    #     out_mv = mat @ vec
    #     assert allclose(out_mv, jnp.zeros_like(out_mv))
    #
    #     B = jnp.ones((shape[1], 4))
    #     out_mm = mat @ B
    #     assert allclose(out_mm, jnp.zeros_like(out_mm))
    #     jax.block_until_ready((dense, vec, out_mv, B, out_mm))

    def test_with_data_accepts_scalar(self):
        mat = brainevent.JITCScalarR((1.5, 0.1, 123), shape=(8, 6))
        updated = mat.with_data(2.0)
        assert allclose(updated.weight, 2.0)
        assert updated.prob == mat.prob
        assert updated.seed == mat.seed
        assert updated.shape == mat.shape



# Copyright 2024 BrainX Ecosystem Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


import brainstate
import braintools
import jax
import jax.numpy as jnp
import numpy as np
import pytest

from brainevent._coo.float import coomv, coomm, coomv_p, coomm_p
from brainevent._coo.test_util import _get_coo, vector_coo, matrix_coo, coo_vector, coo_matrix

PLATFORM = jax.default_backend()
COOMV_IMPLEMENTATIONS = tuple(coomv_p.available_backends(PLATFORM))
COOMM_IMPLEMENTATIONS = tuple(coomm_p.available_backends(PLATFORM))

if not COOMV_IMPLEMENTATIONS:
    pytest.skip(f'No coomv implementation on platform={PLATFORM}', allow_module_level=True)
if not COOMM_IMPLEMENTATIONS:
    pytest.skip(f'No coomm implementation on platform={PLATFORM}', allow_module_level=True)


class TestVectorCOO:
    @pytest.mark.parametrize('implementation', COOMV_IMPLEMENTATIONS)
    @pytest.mark.parametrize('homo_w', [True, False])
    def test_vector_coo(self, implementation, homo_w):
        m, n = 20, 40
        x = brainstate.random.rand(m)
        row, col = _get_coo(m, n, 0.1)

        data = 1.5 if homo_w else braintools.init.Normal(0., 1.)(row.shape)
        y = coomv(data, row, col, x, shape=(m, n), transpose=True, backend=implementation)
        y2 = vector_coo(x, data, row, col, (m, n))
        assert jnp.allclose(y, y2, rtol=1e-3, atol=1e-3)
        jax.block_until_ready((x, row, col, y, y2))

    @pytest.mark.parametrize('implementation', COOMV_IMPLEMENTATIONS)
    @pytest.mark.parametrize('homo_w', [True, False])
    def test_coo_vector(self, implementation, homo_w):
        m, n = 20, 40
        v = brainstate.random.rand(n)
        row, col = _get_coo(m, n, 0.1)

        data = 1.5 if homo_w else braintools.init.Normal(0., 1.)(row.shape)
        y = coomv(data, row, col, v, shape=(m, n), transpose=False, backend=implementation)
        y2 = coo_vector(v, data, row, col, (m, n))
        assert jnp.allclose(y, y2, rtol=1e-3, atol=1e-3)
        jax.block_until_ready((v, row, col, y, y2))

    @pytest.mark.parametrize('implementation', COOMV_IMPLEMENTATIONS)
    @pytest.mark.parametrize('homo_w', [True, False])
    def test_vector_coo_vmap_vector(self, implementation, homo_w):
        n_batch, m, n = 10, 20, 40
        xs = brainstate.random.rand(n_batch, m)
        row, col = _get_coo(m, n, 0.1)

        data = 1.5 if homo_w else braintools.init.Normal(0., 1.)(row.shape)
        y = jax.vmap(
            lambda x: coomv(data, row, col, x, shape=(m, n), transpose=True, backend=implementation)
        )(xs)
        y2 = jax.vmap(lambda x: vector_coo(x, data, row, col, (m, n)))(xs)

        assert jnp.allclose(y, y2, rtol=1e-3, atol=1e-3)
        jax.block_until_ready((xs, row, col, y, y2))

    @pytest.mark.parametrize('implementation', COOMV_IMPLEMENTATIONS)
    @pytest.mark.parametrize('homo_w', [True, False])
    @pytest.mark.parametrize('transpose', [True, False])
    def test_coomv_single_nnz(self, implementation, homo_w, transpose):
        m, n = 10, 20
        row = np.array([3], dtype=np.int32)
        col = np.array([7], dtype=np.int32)
        x = jnp.zeros(m if transpose else n, dtype=jnp.float32)
        x = x.at[3 if transpose else 7].set(1.0)
        data = 2.5 if homo_w else jnp.array([2.5])
        y = coomv(data, row, col, x, shape=(m, n), transpose=transpose, backend=implementation)
        y2 = vector_coo(x, data, row, col, (m, n)) if transpose else coo_vector(x, data, row, col, (m, n))
        assert jax.block_until_ready(jnp.allclose(y, y2, atol=1e-6))

    @pytest.mark.parametrize('implementation', COOMV_IMPLEMENTATIONS)
    @pytest.mark.parametrize('transpose', [True, False])
    def test_coomv_empty(self, implementation, transpose):
        m, n = 10, 20
        row = np.array([], dtype=np.int32)
        col = np.array([], dtype=np.int32)
        x = jnp.ones(m if transpose else n, dtype=jnp.float32)
        data = jnp.array([1.0])
        y = coomv(data, row, col, x, shape=(m, n), transpose=transpose, backend=implementation)
        expected_size = n if transpose else m
        assert jax.block_until_ready(jnp.allclose(y, jnp.zeros(expected_size)))

    @pytest.mark.parametrize('implementation', COOMV_IMPLEMENTATIONS)
    @pytest.mark.parametrize('nnz', [32, 33])
    @pytest.mark.parametrize('homo_w', [True, False])
    def test_coomv_block_boundary(self, implementation, nnz, homo_w):
        m, n = 100, 200
        rng = np.random.default_rng(42)
        row = rng.integers(0, m, size=nnz, dtype=np.int32)
        col = rng.integers(0, n, size=nnz, dtype=np.int32)
        x = jnp.asarray(rng.standard_normal(n), dtype=jnp.float32)
        data = 1.5 if homo_w else jnp.asarray(rng.standard_normal(nnz), dtype=jnp.float32)
        y = coomv(data, row, col, x, shape=(m, n), transpose=False, backend=implementation)
        y2 = coo_vector(x, data, row, col, (m, n))
        assert jax.block_until_ready(jnp.allclose(y, y2, rtol=1e-5, atol=1e-5))

    @pytest.mark.parametrize('implementation', COOMV_IMPLEMENTATIONS)
    @pytest.mark.parametrize('homo_w', [True, False])
    def test_coomv_heavy_duplicates(self, implementation, homo_w):
        m, n, nnz = 5, 5, 200
        row = np.zeros(nnz, dtype=np.int32)
        col = np.zeros(nnz, dtype=np.int32)
        x = jnp.ones(n, dtype=jnp.float32)
        data = 1.0 if homo_w else jnp.ones(nnz, dtype=jnp.float32)
        y = coomv(data, row, col, x, shape=(m, n), transpose=False, backend=implementation)
        y2 = coo_vector(x, data, row, col, (m, n))
        assert jax.block_until_ready(jnp.allclose(y, y2, rtol=1e-5, atol=1e-5))

    @pytest.mark.parametrize('implementation', COOMV_IMPLEMENTATIONS)
    @pytest.mark.parametrize('shape', [(10000, 5), (5, 10000)])
    @pytest.mark.parametrize('homo_w', [True, False])
    def test_coomv_extreme_shapes(self, implementation, shape, homo_w):
        m, n = shape
        nnz = 50
        rng = np.random.default_rng(123)
        row = rng.integers(0, m, size=nnz, dtype=np.int32)
        col = rng.integers(0, n, size=nnz, dtype=np.int32)
        x = jnp.asarray(rng.standard_normal(n), dtype=jnp.float32)
        data = 1.5 if homo_w else jnp.asarray(rng.standard_normal(nnz), dtype=jnp.float32)
        y = coomv(data, row, col, x, shape=(m, n), transpose=False, backend=implementation)
        y2 = coo_vector(x, data, row, col, (m, n))
        assert jax.block_until_ready(jnp.allclose(y, y2, rtol=1e-5, atol=1e-5))

    @pytest.mark.parametrize('implementation', COOMV_IMPLEMENTATIONS)
    @pytest.mark.parametrize('dtype', [jnp.float32, jnp.float64])
    @pytest.mark.parametrize('homo_w', [True, False])
    def test_coomv_dtype(self, implementation, dtype, homo_w):
        with brainstate.environ.context(precision=64 if dtype == jnp.float64 else 32):
            m, n = 30, 50
            rng = np.random.default_rng(99)
            row = rng.integers(0, m, size=60, dtype=np.int32)
            col = rng.integers(0, n, size=60, dtype=np.int32)
            x = jnp.asarray(rng.standard_normal(n), dtype=dtype)
            data = jnp.asarray(1.5, dtype=dtype) if homo_w else jnp.asarray(rng.standard_normal(60), dtype=dtype)
            y = coomv(data, row, col, x, shape=(m, n), transpose=False, backend=implementation)
            y2 = coo_vector(x, data, row, col, (m, n))
            assert y.dtype == dtype
            assert jax.block_until_ready(jnp.allclose(y, y2, rtol=1e-5, atol=1e-5))

    def _test_vjp(self, implementation, homo_w, transpose):
        n_in = 20
        n_out = 30
        shape = (n_in, n_out)
        x = brainstate.random.rand(n_in) if transpose else brainstate.random.rand(n_out)

        row, col = _get_coo(n_in, n_out, 0.2)
        w = 1.5 if homo_w else braintools.init.Normal(0., 1.)(row.shape)

        def f_api(x, w):
            if transpose:
                r = coomv(w, row, col, x, shape=shape, transpose=True, backend=implementation)
            else:
                r = coomv(w, row, col, x, shape=shape, transpose=False, backend=implementation)
            return r.sum()

        r = jax.grad(f_api, argnums=(0, 1))(x, w)

        def f_jax(x, w):
            if transpose:
                r = vector_coo(x, w, row, col, shape=shape)
            else:
                r = coo_vector(x, w, row, col, shape=shape)
            return r.sum()

        r2 = jax.grad(f_jax, argnums=(0, 1))(x, w)
        assert jnp.allclose(r[0], r2[0], rtol=1e-3, atol=1e-3)
        assert jnp.allclose(r[1], r2[1], rtol=1e-3, atol=1e-3)
        jax.block_until_ready((x, row, col, r, r2))

    @pytest.mark.parametrize('implementation', COOMV_IMPLEMENTATIONS)
    @pytest.mark.parametrize('homo_w', [True, False])
    @pytest.mark.parametrize('transpose', [True, False])
    def test_vjp(self, implementation, transpose, homo_w):
        self._test_vjp(
            implementation=implementation,
            homo_w=homo_w,

            transpose=transpose,
        )

    def _test_jvp(self, implementation, homo_w, transpose):
        n_in = 20
        n_out = 30
        shape = (n_in, n_out)
        x = brainstate.random.rand(n_in if transpose else n_out)
        row, col = _get_coo(n_in, n_out, 0.1)

        w = 1.5 if homo_w else braintools.init.Normal(0., 1.)(row.shape)

        def f_api(x, w):
            if transpose:
                r = coomv(w, row, col, x, shape=shape, transpose=True, backend=implementation)
            else:
                r = coomv(w, row, col, x, shape=shape, transpose=False, backend=implementation)
            return r

        o1, r1 = jax.jvp(f_api, (x, w), (jnp.ones_like(x), jnp.ones_like(w)))

        def f_jax(x, w):
            if transpose:
                r = vector_coo(x, w, row, col, shape=shape)
            else:
                r = coo_vector(x, w, row, col, shape=shape)
            return r

        o2, r2 = jax.jvp(f_jax, (x, w), (jnp.ones_like(x), jnp.ones_like(w)))
        assert jnp.allclose(r1, r2, rtol=1e-3, atol=1e-3)
        assert jnp.allclose(o1, o2, rtol=1e-3, atol=1e-3)
        jax.block_until_ready((x, row, col, o1, r1, o2, r2))

    @pytest.mark.parametrize('implementation', COOMV_IMPLEMENTATIONS)
    @pytest.mark.parametrize('homo_w', [True, False])
    @pytest.mark.parametrize('transpose', [True, False])
    def test_jvp(self, implementation, transpose, homo_w):
        self._test_jvp(
            implementation=implementation,
            homo_w=homo_w,
            transpose=transpose,
        )


class TestMatrixCOO:
    @pytest.mark.parametrize('implementation', COOMM_IMPLEMENTATIONS)
    @pytest.mark.parametrize('homo_w', [True, False])
    def test_matrix_coo(self, implementation, homo_w):
        k, m, n = 10, 20, 40
        x = brainstate.random.rand(k, m)
        row, col = _get_coo(m, n, 0.1)

        data = 1.5 if homo_w else braintools.init.Normal(0., 1.)(row.shape)
        y = coomm(data, row, col, x.T, shape=(m, n), transpose=True, backend=implementation).T
        y2 = matrix_coo(x, data, row, col, (m, n))
        assert jnp.allclose(y, y2, rtol=1e-3, atol=1e-3)
        jax.block_until_ready((x, row, col, y, y2))

    @pytest.mark.parametrize('implementation', COOMM_IMPLEMENTATIONS)
    @pytest.mark.parametrize('homo_w', [True, False])
    def test_coo_matrix(self, implementation, homo_w):
        m, n, k = 20, 40, 10
        x = brainstate.random.rand(n, k)
        row, col = _get_coo(m, n, 0.1)

        data = 1.5 if homo_w else braintools.init.Normal(0., 1.)(row.shape)
        y = coomm(data, row, col, x, shape=(m, n), transpose=False, backend=implementation)
        y2 = coo_matrix(x, data, row, col, (m, n))
        assert jnp.allclose(y, y2, rtol=1e-3, atol=1e-3)
        jax.block_until_ready((x, row, col, y, y2))



# Copyright 2024 BrainX Ecosystem Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import os

os.environ['JAX_TRACEBACK_FILTERING'] = 'off'

import brainstate
import braintools
import jax
import jax.numpy as jnp
import pytest

from brainevent._csr.float import csrmv, csrmv_p, csrmm, csrmm_p
from brainevent._csr.test_util import get_csr, vector_csr, matrix_csr, csr_vector, csr_matrix

platform = jax.default_backend()
CSRMV_IMPLEMENTATIONS = tuple(csrmv_p.available_backends(platform))
CSRMM_IMPLEMENTATIONS = tuple(csrmm_p.available_backends(platform))


def _make_data(homo_w, shape):
    if homo_w:
        return jnp.asarray(1.5, dtype=jnp.float32)
    return braintools.init.Normal(0.0, 1.0)(shape)


def _vector_csr_api(x, data, indices, indptr, shape, implementation):
    return csrmv(
        data,
        indices,
        indptr,
        x,
        shape=shape,
        transpose=True,
        backend=implementation,
    )


def _csr_vector_api(x, data, indices, indptr, shape, implementation):
    return csrmv(
        data,
        indices,
        indptr,
        x,
        shape=shape,
        transpose=False,
        backend=implementation,
    )


def _matrix_csr_api(x, data, indices, indptr, shape, implementation):
    return csrmm(
        data,
        indices,
        indptr,
        x.T,
        shape=shape,
        transpose=True,
        backend=implementation,
    ).T


def _csr_matrix_api(x, data, indices, indptr, shape, implementation):
    return csrmm(
        data,
        indices,
        indptr,
        x,
        shape=shape,
        transpose=False,
        backend=implementation,
    )


def _row_ids_from_indptr(indptr):
    indptr = jnp.asarray(indptr)
    counts = jnp.diff(indptr)
    return jnp.repeat(jnp.arange(counts.shape[0], dtype=indptr.dtype), counts)


@pytest.mark.skipif(
    not CSRMV_IMPLEMENTATIONS,
    reason=f'No csrmv implementation on platform={platform}',
)
class TestFloatCSRMV:
    @pytest.mark.parametrize('implementation', CSRMV_IMPLEMENTATIONS)
    @pytest.mark.parametrize('homo_w', [True, False])
    def test_vector_csr(self, implementation, homo_w):
        m, n = 20, 40
        x = brainstate.random.rand(m)
        indptr, indices = get_csr(m, n, 0.1)
        data = _make_data(homo_w, indices.shape)

        y = _vector_csr_api(x, data, indices, indptr, (m, n), implementation)
        y_ref = vector_csr(x, data, indices, indptr, (m, n))
        assert jnp.allclose(y, y_ref, rtol=1e-3, atol=1e-3)

        jax.block_until_ready((x, indptr, indices, data, y, y_ref))

    @pytest.mark.parametrize('implementation', CSRMV_IMPLEMENTATIONS)
    @pytest.mark.parametrize('homo_w', [True, False])
    def test_csr_vector(self, implementation, homo_w):
        m, n = 20, 40
        v = brainstate.random.rand(n)
        indptr, indices = get_csr(m, n, 0.1)
        data = _make_data(homo_w, indices.shape)

        y = _csr_vector_api(v, data, indices, indptr, (m, n), implementation)
        y_ref = csr_vector(v, data, indices, indptr, (m, n))
        assert jnp.allclose(y, y_ref, rtol=1e-3, atol=1e-3)

        jax.block_until_ready((v, indptr, indices, data, y, y_ref))

    @pytest.mark.parametrize('implementation', CSRMV_IMPLEMENTATIONS)
    @pytest.mark.parametrize('homo_w', [True, False])
    def test_vector_csr_vmap_vector(self, implementation, homo_w):
        n_batch, m, n = 10, 20, 40
        xs = brainstate.random.rand(n_batch, m)
        indptr, indices = get_csr(m, n, 0.1)
        data = _make_data(homo_w, indices.shape)

        y = brainstate.transform.vmap2(
            lambda x: _vector_csr_api(x, data, indices, indptr, (m, n), implementation)
        )(xs)
        y_ref = brainstate.transform.vmap2(lambda x: vector_csr(x, data, indices, indptr, (m, n)))(xs)
        assert jnp.allclose(y, y_ref, rtol=1e-3, atol=1e-3)

        jax.block_until_ready((xs, indptr, indices, data, y, y_ref))

    @pytest.mark.parametrize('implementation', CSRMV_IMPLEMENTATIONS)
    @pytest.mark.parametrize('homo_w', [True, False])
    @pytest.mark.parametrize('replace', [True, False])
    @pytest.mark.parametrize('transpose', [True, False])
    def test_vjp(self, implementation, homo_w, replace, transpose):
        n_in = 20
        n_out = 30
        shape = (n_in, n_out)
        x = brainstate.random.rand(n_in) if transpose else brainstate.random.rand(n_out)

        indptr, indices = get_csr(n_in, n_out, 0.2, replace=replace)
        w = _make_data(homo_w, indices.shape)

        def f_api(x_arg, w_arg):
            if transpose:
                return _vector_csr_api(x_arg, w_arg, indices, indptr, shape, implementation).sum()
            return _csr_vector_api(x_arg, w_arg, indices, indptr, shape, implementation).sum()

        def f_ref(x_arg, w_arg):
            if transpose:
                return vector_csr(x_arg, w_arg, indices, indptr, shape=shape).sum()
            return csr_vector(x_arg, w_arg, indices, indptr, shape=shape).sum()

        r = jax.grad(f_api, argnums=(0, 1))(x, w)
        r_ref = jax.grad(f_ref, argnums=(0, 1))(x, w)
        assert jnp.allclose(r[0], r_ref[0], rtol=1e-3, atol=1e-3)
        assert jnp.allclose(r[1], r_ref[1], rtol=1e-3, atol=1e-3)

        jax.block_until_ready((x, indptr, indices, w, r, r_ref))

    @pytest.mark.parametrize('implementation', CSRMV_IMPLEMENTATIONS)
    @pytest.mark.parametrize('homo_w', [True, False])
    @pytest.mark.parametrize('replace', [True, False])
    @pytest.mark.parametrize('transpose', [True, False])
    def test_jvp(self, implementation, homo_w, replace, transpose):
        n_in = 20
        n_out = 30
        shape = (n_in, n_out)
        x = brainstate.random.rand(n_in if transpose else n_out)

        indptr, indices = get_csr(n_in, n_out, 0.1, replace=replace)
        w = _make_data(homo_w, indices.shape)

        def f_api(x_arg, w_arg):
            if transpose:
                return _vector_csr_api(x_arg, w_arg, indices, indptr, shape, implementation)
            return _csr_vector_api(x_arg, w_arg, indices, indptr, shape, implementation)

        def f_ref(x_arg, w_arg):
            if transpose:
                return vector_csr(x_arg, w_arg, indices, indptr, shape=shape)
            return csr_vector(x_arg, w_arg, indices, indptr, shape=shape)

        o1, r1 = jax.jvp(f_api, (x, w), (jnp.ones_like(x), jnp.ones_like(w)))
        o2, r2 = jax.jvp(f_ref, (x, w), (jnp.ones_like(x), jnp.ones_like(w)))
        assert jnp.allclose(r1, r2, rtol=1e-3, atol=1e-3)
        assert jnp.allclose(o1, o2, rtol=1e-3, atol=1e-3)

        jax.block_until_ready((x, indptr, indices, w, o1, r1, o2, r2))


@pytest.mark.skipif(
    not CSRMM_IMPLEMENTATIONS,
    reason=f'No csrmm implementation on platform={platform}',
)
class TestFloatCSRMM:
    @pytest.mark.parametrize('implementation', CSRMM_IMPLEMENTATIONS)
    @pytest.mark.parametrize('homo_w', [True, False])
    def test_matrix_csr(self, implementation, homo_w):
        k, m, n = 10, 20, 40
        x = brainstate.random.rand(k, m)
        indptr, indices = get_csr(m, n, 0.1)
        data = _make_data(homo_w, indices.shape)

        y = _matrix_csr_api(x, data, indices, indptr, (m, n), implementation)
        y_ref = matrix_csr(x, data, indices, indptr, (m, n))
        assert jnp.allclose(y, y_ref, rtol=1e-3, atol=1e-3)

        jax.block_until_ready((x, indptr, indices, data, y, y_ref))

    @pytest.mark.parametrize('implementation', CSRMM_IMPLEMENTATIONS)
    @pytest.mark.parametrize('homo_w', [True, False])
    def test_csr_matrix(self, implementation, homo_w):
        m, n, k = 20, 40, 10
        x = brainstate.random.rand(n, k)
        indptr, indices = get_csr(m, n, 0.1)
        data = _make_data(homo_w, indices.shape)

        y = _csr_matrix_api(x, data, indices, indptr, (m, n), implementation)
        y_ref = csr_matrix(x, data, indices, indptr, (m, n))
        assert jnp.allclose(y, y_ref, rtol=1e-3, atol=1e-3)

        jax.block_until_ready((x, indptr, indices, data, y, y_ref))

    @pytest.mark.parametrize('implementation', CSRMM_IMPLEMENTATIONS)
    @pytest.mark.parametrize('homo_w', [True, False])
    @pytest.mark.parametrize('replace', [True, False])
    @pytest.mark.parametrize('transpose', [True, False])
    def test_vjp(self, implementation, homo_w, replace, transpose):
        m, n, k = 20, 30, 8
        shape = (m, n)
        x = brainstate.random.rand(m, k) if transpose else brainstate.random.rand(n, k)

        indptr, indices = get_csr(m, n, 0.2, replace=replace)
        w = _make_data(homo_w, indices.shape)

        def f_api(x_arg, w_arg):
            return csrmm(
                w_arg,
                indices,
                indptr,
                x_arg,
                shape=shape,
                transpose=transpose,
                backend=implementation,
            ).sum()

        def f_ref(x_arg, w_arg):
            if transpose:
                return matrix_csr(x_arg.T, w_arg, indices, indptr, shape).T.sum()
            return csr_matrix(x_arg, w_arg, indices, indptr, shape).sum()

        r = jax.grad(f_api, argnums=(0, 1))(x, w)
        r_ref = jax.grad(f_ref, argnums=(0, 1))(x, w)
        assert jnp.allclose(r[0], r_ref[0], rtol=1e-3, atol=1e-3)
        assert jnp.allclose(r[1], r_ref[1], rtol=1e-3, atol=1e-3)

        jax.block_until_ready((x, indptr, indices, w, r, r_ref))

    @pytest.mark.parametrize('implementation', CSRMM_IMPLEMENTATIONS)
    @pytest.mark.parametrize('homo_w', [True, False])
    @pytest.mark.parametrize('replace', [True, False])
    @pytest.mark.parametrize('transpose', [True, False])
    def test_jvp(self, implementation, homo_w, replace, transpose):
        m, n, k = 20, 30, 8
        shape = (m, n)
        x = brainstate.random.rand(m, k) if transpose else brainstate.random.rand(n, k)

        indptr, indices = get_csr(m, n, 0.1, replace=replace)
        w = _make_data(homo_w, indices.shape)

        def f_api(x_arg, w_arg):
            return csrmm(
                w_arg,
                indices,
                indptr,
                x_arg,
                shape=shape,
                transpose=transpose,
                backend=implementation,
            )

        def f_ref(x_arg, w_arg):
            if transpose:
                return matrix_csr(x_arg.T, w_arg, indices, indptr, shape).T
            return csr_matrix(x_arg, w_arg, indices, indptr, shape)

        o1, r1 = jax.jvp(f_api, (x, w), (jnp.ones_like(x), jnp.ones_like(w)))
        o2, r2 = jax.jvp(f_ref, (x, w), (jnp.ones_like(x), jnp.ones_like(w)))
        assert jnp.allclose(r1, r2, rtol=1e-3, atol=1e-3)
        assert jnp.allclose(o1, o2, rtol=1e-3, atol=1e-3)

        jax.block_until_ready((x, indptr, indices, w, o1, r1, o2, r2))



# Copyright 2025 BrainX Ecosystem Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

# -*- coding: utf-8 -*-


import brainstate
import braintools
import jax
import jax.numpy as jnp
import numpy as np
import pytest

import brainevent
from brainevent._coo.test_util import _get_coo, vector_coo, matrix_coo, coo_vector, coo_matrix


def gen_sparse_matrix(shape, prob=0.2):
    matrix = np.random.rand(*shape)
    matrix = np.where(matrix < prob, matrix, 0.)
    return jnp.asarray(matrix, dtype=float)


class TestCOOAddSub:
    @pytest.mark.parametrize('shape', [(20, 30), (100, 50)])
    def test_add_scalar(self, shape):
        matrix = gen_sparse_matrix(shape)
        coo = brainevent.COO.fromdense(matrix)
        result = coo + 5.0
        expected = coo.todense() + 5.0
        assert not isinstance(result, brainevent.COO)
        assert jnp.allclose(result, expected)
        jax.block_until_ready((matrix, result, expected))

    @pytest.mark.parametrize('shape', [(20, 30), (100, 50)])
    def test_radd_scalar(self, shape):
        matrix = gen_sparse_matrix(shape)
        coo = brainevent.COO.fromdense(matrix)
        result = 5.0 + coo
        expected = 5.0 + coo.todense()
        assert not isinstance(result, brainevent.COO)
        assert jnp.allclose(result, expected)
        jax.block_until_ready((matrix, result, expected))

    @pytest.mark.parametrize('shape', [(20, 30), (100, 50)])
    def test_sub_dense(self, shape):
        matrix = gen_sparse_matrix(shape)
        coo = brainevent.COO.fromdense(matrix)
        ones = jnp.ones(shape)
        result = coo - ones
        expected = coo.todense() - ones
        assert not isinstance(result, brainevent.COO)
        assert jnp.allclose(result, expected)
        jax.block_until_ready((matrix, ones, result, expected))

    @pytest.mark.parametrize('shape', [(20, 30), (100, 50)])
    def test_rsub_dense(self, shape):
        matrix = gen_sparse_matrix(shape)
        coo = brainevent.COO.fromdense(matrix)
        ones = jnp.ones(shape)
        result = ones - coo
        expected = ones - coo.todense()
        assert not isinstance(result, brainevent.COO)
        assert jnp.allclose(result, expected)
        jax.block_until_ready((matrix, ones, result, expected))

    @pytest.mark.parametrize('shape', [(20, 30), (100, 50)])
    def test_mul_scalar_stays_sparse(self, shape):
        matrix = gen_sparse_matrix(shape)
        coo = brainevent.COO.fromdense(matrix)
        result = coo * 2.0
        assert isinstance(result, brainevent.COO)
        expected = coo.todense() * 2.0
        assert jnp.allclose(result.todense(), expected)
        jax.block_until_ready((matrix, expected))


class TestCOOToCsr:
    @pytest.mark.parametrize('shape', [(20, 30), (100, 50)])
    def test_tocsr(self, shape):
        matrix = gen_sparse_matrix(shape)
        coo = brainevent.COO.fromdense(matrix)
        csr = coo.tocsr()
        assert isinstance(csr, brainevent.CSR)
        assert jnp.allclose(csr.todense(), coo.todense())
        jax.block_until_ready((matrix,))


class TestCOOMatmul:
    """Test COO.__matmul__ and COO.__rmatmul__ operator dispatch."""

    @pytest.mark.parametrize('shape', [(20, 30), (100, 50)])
    def test_matmul_binary_vector(self, shape):
        matrix = gen_sparse_matrix(shape)
        coo = brainevent.COO.fromdense(matrix)
        v = brainevent.BinaryArray(brainstate.random.rand(shape[1]) < 0.5)
        result = coo @ v
        expected = matrix @ v.value.astype(float)
        assert jnp.allclose(result, expected)
        jax.block_until_ready((matrix, result, expected))

    @pytest.mark.parametrize('shape', [(20, 30), (100, 50)])
    def test_rmatmul_binary_vector(self, shape):
        matrix = gen_sparse_matrix(shape)
        coo = brainevent.COO.fromdense(matrix)
        v = brainevent.BinaryArray(brainstate.random.rand(shape[0]) < 0.5)
        result = v @ coo
        expected = v.value.astype(float) @ matrix
        assert jnp.allclose(result, expected)
        jax.block_until_ready((matrix, result, expected))

    @pytest.mark.parametrize('shape', [(20, 30), (100, 50)])
    def test_matmul_binary_matrix(self, shape):
        matrix = gen_sparse_matrix(shape)
        coo = brainevent.COO.fromdense(matrix)
        B = brainevent.BinaryArray(brainstate.random.rand(shape[1], 10) < 0.5)
        result = coo @ B
        expected = matrix @ B.value.astype(float)
        assert jnp.allclose(result, expected, rtol=1e-3, atol=1e-3)
        jax.block_until_ready((matrix, result, expected))

    @pytest.mark.parametrize('shape', [(20, 30), (100, 50)])
    def test_rmatmul_binary_matrix(self, shape):
        matrix = gen_sparse_matrix(shape)
        coo = brainevent.COO.fromdense(matrix)
        B = brainevent.BinaryArray(brainstate.random.rand(10, shape[0]) < 0.5)
        result = B @ coo
        expected = B.value.astype(float) @ matrix
        assert jnp.allclose(result, expected, rtol=1e-3, atol=1e-3)
        jax.block_until_ready((matrix, result, expected))

    @pytest.mark.parametrize('shape', [(20, 30), (100, 50)])
    def test_matmul_dense_vector(self, shape):
        matrix = gen_sparse_matrix(shape)
        coo = brainevent.COO.fromdense(matrix)
        v = jnp.asarray(np.random.rand(shape[1]))
        result = coo @ v
        expected = matrix @ v
        assert jnp.allclose(result, expected, rtol=1e-3, atol=1e-3)
        jax.block_until_ready((matrix, v, result, expected))

    @pytest.mark.parametrize('shape', [(20, 30), (100, 50)])
    def test_rmatmul_dense_vector(self, shape):
        matrix = gen_sparse_matrix(shape)
        coo = brainevent.COO.fromdense(matrix)
        v = jnp.asarray(np.random.rand(shape[0]))
        result = v @ coo
        expected = v @ matrix
        assert jnp.allclose(result, expected, rtol=1e-3, atol=1e-3)
        jax.block_until_ready((matrix, v, result, expected))


class TestCOOBinaryOperator:
    def test_event_homo_bool(self):
        for dat in [1., 2., 3.]:
            mask = (brainstate.random.rand(10, 20) < 0.1).astype(float) * dat
            coo = brainevent.COO.fromdense(mask)

            v = brainevent.BinaryArray(brainstate.random.rand(20) < 0.5)
            assert jnp.allclose(
                mask.astype(float) @ v.value.astype(float),
                coo @ v
            )

            v = brainevent.BinaryArray(brainstate.random.rand(10) < 0.5)
            assert jnp.allclose(
                v.value.astype(float) @ mask.astype(float),
                v @ coo
            )
            jax.block_until_ready((mask,))

    def test_event_homo_float_as_bool(self):
        mat = brainstate.random.rand(10, 20)
        mask = (mat < 0.1).astype(float) * mat
        coo = brainevent.COO.fromdense(mask)

        v = brainevent.BinaryArray((brainstate.random.rand(20) < 0.5).astype(float))
        assert jnp.allclose(
            mask.astype(float) @ v.value.astype(float),
            coo @ v
        )

        v = brainevent.BinaryArray((brainstate.random.rand(10) < 0.5).astype(float))
        assert jnp.allclose(
            v.value.astype(float) @ mask.astype(float),
            v @ coo
        )
        jax.block_until_ready((mat, mask))

    def test_event_homo_other(self):
        mat = brainstate.random.rand(10, 20)
        mask = (brainstate.random.rand(10, 20) < 0.1) * mat
        coo = brainevent.COO.fromdense(mask)

        v = brainevent.BinaryArray(brainstate.random.rand(20) < 0.5)
        assert jnp.allclose(
            mask.astype(float) @ v.value.astype(float),
            coo @ v
        )

        v = brainevent.BinaryArray(brainstate.random.rand(10) < 0.5)
        assert jnp.allclose(
            v.value.astype(float) @ mask.astype(float),
            v @ coo
        )
        jax.block_until_ready((mat, mask))


class TestCOOBinaryVectorOperator:
    @pytest.mark.parametrize('homo_w', [True, False])
    def test_vector_coo(self, homo_w):
        m, n = 20, 40
        x = brainstate.random.rand(m) < 0.1
        row, col = _get_coo(m, n, 0.1)

        data = 1.5 if homo_w else braintools.init.Normal(0., 1.)(row.shape)
        coo = brainevent.COO([data, row, col], shape=(m, n))
        y = brainevent.BinaryArray(x) @ coo
        y2 = vector_coo(x, coo.data, row, col, (m, n))
        assert jnp.allclose(y, y2, rtol=1e-5, atol=1e-5)
        jax.block_until_ready((x, row, col, y, y2))

    @pytest.mark.parametrize('homo_w', [True, False])
    def test_vector_coo_vmap_vector(self, homo_w):
        n_batch, m, n = 10, 20, 40
        xs = brainstate.random.rand(n_batch, m) < 0.1
        row, col = _get_coo(m, n, 0.1)

        data = 1.5 if homo_w else braintools.init.Normal(0., 1.)(row.shape)
        coo = brainevent.COO([data, row, col], shape=(m, n))
        y = jax.vmap(lambda x: brainevent.BinaryArray(x) @ coo)(xs)
        y2 = jax.vmap(lambda x: vector_coo(x, coo.data, row, col, (m, n)))(xs)
        assert jnp.allclose(y, y2, rtol=1e-3, atol=1e-3)
        jax.block_until_ready((xs, row, col, y, y2))

    @pytest.mark.parametrize('homo_w', [True, False])
    def test_coo_vector(self, homo_w):
        m, n = 20, 40
        v = brainstate.random.rand(n) < 0.1
        row, col = _get_coo(m, n, 0.2)

        data = 1.5 if homo_w else braintools.init.Normal(0., 1.)(row.shape)
        coo = brainevent.COO([data, row, col], shape=(m, n))
        y = coo @ brainevent.BinaryArray(v)
        y2 = coo_vector(v, coo.data, row, col, (m, n))
        assert jnp.allclose(y, y2, rtol=1e-5, atol=1e-5)
        jax.block_until_ready((v, row, col, y, y2))

    def _test_vjp(self, homo_w, replace, transpose):
        n_in = 20
        n_out = 30
        shape = (n_in, n_out)
        x = brainstate.random.rand(n_in) if transpose else brainstate.random.rand(n_out)
        x = (x < 0.6).astype(float)

        row, col = _get_coo(n_in, n_out, 0.2, replace=replace)
        w = 1.5 if homo_w else braintools.init.Normal(0., 1.)(row.shape)
        coo = brainevent.COO((w, row, col), shape=shape)

        def f_brainevent(x, w):
            if transpose:
                r = brainevent.BinaryArray(x) @ coo.with_data(w)
            else:
                r = coo.with_data(w) @ brainevent.BinaryArray(x)
            return r.sum()

        r = jax.grad(f_brainevent, argnums=(0, 1))(x, w)

        def f_jax(x, w):
            if transpose:
                r = vector_coo(x, w, row, col, shape=shape)
            else:
                r = coo_vector(x, w, row, col, shape=shape)
            return r.sum()

        r2 = jax.grad(f_jax, argnums=(0, 1))(x, w)
        assert jnp.allclose(r[0], r2[0], rtol=1e-3, atol=1e-3)
        assert jnp.allclose(r[1], r2[1], rtol=1e-3, atol=1e-3)
        jax.block_until_ready((x, row, col, r, r2))

    @pytest.mark.parametrize('homo_w', [True, False])
    @pytest.mark.parametrize('replace', [True, False])
    @pytest.mark.parametrize('transpose', [True, False])
    def test_vjp(self, transpose, replace, homo_w):
        self._test_vjp(homo_w=homo_w, replace=replace, transpose=transpose)

    def _test_jvp(self, homo_w, replace, transpose):
        n_in = 20
        n_out = 30
        shape = (n_in, n_out)
        x = brainstate.random.rand(n_in if transpose else n_out)
        x = (x < 0.6).astype(float)

        row, col = _get_coo(n_in, n_out, 0.1, replace=replace)
        w = 1.5 if homo_w else braintools.init.Normal(0., 1.)(row.shape)
        coo = brainevent.COO((w, row, col), shape=shape)

        def f_brainevent(x, w):
            if transpose:
                r = brainevent.BinaryArray(x) @ coo.with_data(w)
            else:
                r = coo.with_data(w) @ brainevent.BinaryArray(x)
            return r

        o1, r1 = jax.jvp(f_brainevent, (x, w), (jnp.ones_like(x), jnp.ones_like(w)))

        def f_jax(x, w):
            if transpose:
                r = vector_coo(x, w, row, col, shape=shape)
            else:
                r = coo_vector(x, w, row, col, shape=shape)
            return r

        o2, r2 = jax.jvp(f_jax, (x, w), (jnp.ones_like(x), jnp.ones_like(w)))
        assert jnp.allclose(r1, r2, rtol=1e-3, atol=1e-3)
        assert jnp.allclose(o1, o2, rtol=1e-3, atol=1e-3)
        jax.block_until_ready((x, row, col, o1, r1, o2, r2))

    @pytest.mark.parametrize('homo_w', [True, False])
    @pytest.mark.parametrize('replace', [True, False])
    @pytest.mark.parametrize('transpose', [True, False])
    def test_jvp(self, transpose, replace, homo_w):
        self._test_jvp(homo_w=homo_w, replace=replace, transpose=transpose)


class TestCOOBinaryMatrixOperator:
    @pytest.mark.parametrize('homo_w', [True, False])
    def test_matrix_coo(self, homo_w):
        k, m, n = 10, 20, 40
        x = brainstate.random.rand(k, m) < 0.1
        row, col = _get_coo(m, n, 0.1)

        data = 1.5 if homo_w else braintools.init.Normal(0., 1.)(row.shape)
        coo = brainevent.COO([data, row, col], shape=(m, n))
        y = brainevent.BinaryArray(x) @ coo
        y2 = matrix_coo(x.astype(float), coo.data, row, col, (m, n))
        assert jnp.allclose(y, y2, rtol=1e-3, atol=1e-3)
        jax.block_until_ready((x, row, col, y, y2))

    @pytest.mark.parametrize('homo_w', [True, False])
    def test_coo_matrix(self, homo_w):
        m, n, k = 20, 40, 10
        matrix = brainstate.random.rand(n, k) < 0.1
        row, col = _get_coo(m, n, 0.1)

        data = 1.5 if homo_w else braintools.init.Normal(0., 1.)(row.shape)
        coo = brainevent.COO([data, row, col], shape=(m, n))
        y = coo @ brainevent.BinaryArray(matrix)
        y2 = coo_matrix(matrix.astype(float), coo.data, row, col, (m, n))
        assert jnp.allclose(y, y2)
        jax.block_until_ready((matrix, row, col, y, y2))
